/* ooGreedy - A greedy method for doing ordering and orienting
 * of contigs. */

#undef VERBOSE_LOG

#include "common.h"
#include "linefile.h"
#include "obscure.h"
#include "errabort.h"
#include "memalloc.h"
#include "portable.h"
#include "hash.h"
#include "fa.h"
#include "dlist.h"
#include "diGraph.h"
#include "localmem.h"
#include "hCommon.h"
#include "psl.h"
#include "ooGreedy.h"

int version = 68;       /* Current version number. */

int minFragSize = 1;            /* Minimum size of fragments we'll accept. */
int maxMapDeviation = 750000;   /* No map deviations further than this allowed. */
int freeMapDeviation = 250000;  /* Map deviations to this point free. */

int maxTailSize = 2000;	/* Maximum non-aligning end. */
int minScore = -500;    /* Don't bother gluing past this. */
int minBacScore = -720; /* Bac ends can be a little worse... */

int bridgedBargeGapSize = 50000;	/* # of N's between connected barges. */
int unbridgedBargeGapSize = 100000;     /* # of N's between disconnected barges. */
int fragGapSize = 100;                  /* # of N's between frags within a barge. */

FILE *logFile;	/* File to write decision steps to. */
FILE *raftPsl = NULL;  /* File to write consistent .psl to. */
char looseFileName[512];  /* Where to write loose overlapping clone pairs. */
char tightFileName[512];  /* Where to write tight overlapping clone pairs. */

void usage()
/* Explain usage and exit. */
{
errAbort(
   "ooGreedy - ordering and orienting of unfinished contigs using\n"
   "a relatively simple, but hopefully effective greedy method\n"
   "usage:\n"
   "      ooGreedy directory\n"
   "The program expects the files self.psl, est.psl, mrna.psl, bacEnds.psl\n"
   "bacEndsPairs to live in the directory.  It will create the files\n"
   "gl.N raft.N graph.N raft.psl ooGreedy.N.gl ooGreedy.log barge.N gold.N \n"
   "where N is the version number of ooGreedy (currently %d).\n"
   "These files describe:\n"
   "  gl.N - Where each fragment is in 'golden path' coordinates\n"
   "         along with connectivity info on fragments.\n"
   "  raft.N - Fragments connected by direct sequence overlap\n"
   "  graph.N - connections of rafts by mRNA and BAC end data\n"
   "  raft.psl - Alignments from self.psl involved in building rafts\n"
   "  ooGreedy.N.gl - same as gl.N but lacking connectivity info\n"
   "  ooGreedy.log - record of decisions program made.\n"
   "  barge.N - where clones in the contig belong.\n"
   "  gold.N - what parts of what fragment to use to build final\n"
   "           'golden path' merged sequence for contig."
   , version);
}


void warnHandler(char *format, va_list args)
/* Default error message handler. */
{
if (format != NULL) 
    {
    fprintf(stderr, "[warning] ");
    fprintf(logFile, "[warning] ");
    vfprintf(stderr, format, args);
    vfprintf(logFile, format, args);
    fprintf(stderr, "\n");
    fprintf(logFile, "\n");
    }
}

void logg(char *format, ...)
/* Print status message to log file. */
{
va_list args;
va_start(args, format);
vfprintf(logFile, format, args);
va_end(args);
}

void status(char *format, ...)
/* Print status message to stdout and to log file. */
{
va_list args;
va_start(args, format);
vfprintf(logFile, format, args);
vfprintf(stdout, format, args);
va_end(args);
}

void raftFree(struct raft **pRaft)
/* Free up memory associated with a raft. */
{
struct raft *raft;
if ((raft = *pRaft) != NULL)
    {
    slFreeList(&raft->fragList);
    freez(pRaft);
    }
}

void raftFreeList(struct raft **pList)
/* Free up a list of rafts. */
{
struct raft *el, *next;

for (el = *pList; el != NULL; el = next)
    {
    next = el->next;
    raftFree(&el);
    }
*pList = NULL;
}

void unlinkFragsFromRafts(struct oogFrag *fragList)
/* NULL out frag->raft for all in list. */
{
struct oogFrag *frag;
for (frag = fragList; frag != NULL; frag = frag->next)
    frag->raft = NULL;
}


int cmpEdgeExtension(const void *va, const void *vb)
/* Compare two edgeExtension to sort by seqQual. */
{
const struct edgeExtension *a = *((struct edgeExtension **)va);
const struct edgeExtension *b = *((struct edgeExtension **)vb);
int diff;
diff = b->seqQual - a->seqQual;
if (diff == 0)
    diff = b->edge->score - a->edge->score;
return diff;
}

int cmpFragDefaultPos(const void *va, const void *vb)
/* Compare two oogFrags to sort by default position . */
{
const struct oogFrag *a = *((struct oogFrag **)va);
const struct oogFrag *b = *((struct oogFrag **)vb);
return a->defaultPos - b->defaultPos;
}

int nextRaftId()
/* Return next available id for raft */
{
static int id = 0;
return ++id;
}

int raftFragCmpOffset(const void *va, const void *vb)
/* Compare two raftFrags to sort by offset . */
{
const struct raftFrag *a = *((struct raftFrag **)va);
const struct raftFrag *b = *((struct raftFrag **)vb);
return a->raftOffset - b->raftOffset;
}

int fragSeqQual(struct oogFrag *frag)
/* Return a number that corresponds to sequence quality
 * as estimated by size and phase. */
{
int sizeFactor = frag->size;
if (sizeFactor > 200000) sizeFactor = 200000;
return frag->clone->phase*50000 + sizeFactor;
}

int raftFragCmpSeqQuality(const void *va, const void *vb)
/* Compare two raftFrags to sort by sequence quality . */
{
const struct raftFrag *a = *((struct raftFrag **)va);
const struct raftFrag *b = *((struct raftFrag **)vb);
return fragSeqQual(b->frag) - fragSeqQual(a->frag);
}

char orientChar(int orient)
/* Return character representing orientation. */
{
if (orient > 0) return '+';
else if (orient < 0) return '-';
else return '.';
}

void dumpRaft(struct raft *raft, FILE *out)
/* Print out info about raft. */
{
struct raftFrag *rf;

fprintf(out, "%d raft %d bases %d frags %d default pos\n", 
	raft->id, raft->end - raft->start, slCount(raft->fragList),
	raft->defaultPos);
for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {

    fprintf(out, "%d %c %s %d\n", rf->raftOffset, 
	orientChar(rf->orientation),
	rf->frag->name, rf->frag->size);
    }
}

boolean fragInRaft(struct raft *raftList, 
	struct oogFrag *frag, struct raft **retRaft, struct raftFrag **retRaftFrag)
/* Return TRUE if fragment is already in a raft.  Return which raft it's 
 * and which raft frag if applicable. */
{
struct raft *raft;
struct raftFrag *rf;

for (raft = raftList; raft != NULL; raft = raft->next)
    {
    for (rf = raft->fragList; rf != NULL; rf = rf->next)
	{
	if (rf->frag == frag)
	    {
	    *retRaft = raft;
	    *retRaftFrag = rf;
	    return TRUE;
	    }
	}
    }
*retRaft = NULL;
*retRaftFrag = NULL;
return FALSE;
}

int oogEdgeCmp(const void *va, const void *vb)
/* Compare two oogEdges for least bad first. */
{
const struct oogEdge *a = *((struct oogEdge **)va);
const struct oogEdge *b = *((struct oogEdge **)vb);
return b->score - a->score;
}

void dumpEdge(struct oogEdge *edge, FILE *out)
/* Print out an edge. */
{
fprintf(out, "edge: %11s %c> %11s %s bad %3d tailA %3d tailB %3d match %5d repMatch %5d score %4d\n", edge->aSide->name,
	((edge->orientation > 0) ? '-' : '~'),
	edge->bSide->name, (edge->aCovers ? "covers " : "extends" ),
	edge->milliBad, edge->tailA, edge->tailB, edge->match, edge->repMatch, edge->score);
}


void dumpMrnaLine(struct mrnaLine *ml, FILE *f)
/* Print out mrnaLine. */
{
struct mrnaConnection *mc;
fprintf(f, "mrnaLine: %s score %d connections %d\n", 
	ml->name, ml->score, slCount(ml->mcList));
for (mc = ml->mcList; mc != NULL; mc = mc->next)
    {
    fprintf(f, "  raft %d %c score %d\n", mc->raft->id, 
	(mc->orientation > 0 ? '+' : '-'),
	mc->score);
    }
}

int mrnaLineCmp(const void *va, const void *vb)
/* Compare to sort based on score - best score first. */
{
const struct mrnaLine *a = *((struct mrnaLine **)va);
const struct mrnaLine *b = *((struct mrnaLine **)vb);
return b->score - a->score;
}


int markBestConnections(struct mrnaConnection *list)
/* Set the best flags for connections that are best for a given raft. 
 * Returns very best score. */
{
struct mrnaConnection *startEl, *el;
int veryBestScore = -0x3fffffff;

startEl = list;
for (;;)
    {
    int bestScore = -0x3fffffff;
    struct raft *curRaft;
    struct mrnaConnection *bestEl;

    while (startEl != NULL && startEl->raftTested)
	startEl = startEl->next;
    if (startEl == NULL)
	break;
    curRaft = startEl->raft;
    for (el = startEl; el != NULL; el = el->next)
	{
	if (el->raft == curRaft)
	    {
	    if (bestScore < el->score)
		{
		bestScore = el->score;
		bestEl = el;
		if (veryBestScore < bestScore)
		    veryBestScore = bestScore;
		}
	    el->raftTested = TRUE;
	    }
	}
    bestEl->bestForRaft = TRUE;
    }
return veryBestScore;
}


struct mrnaConnection *cleanMrnaConnectionList(struct mrnaConnection *oldList)
/* Clean up old list, removing crufty connections.  */
{
struct mrnaConnection *mc, *next;
struct mrnaConnection *newList = NULL;
int veryBestScore;
int threshold;
int score;

veryBestScore = markBestConnections(oldList);
if (veryBestScore < -100)
    return NULL;
threshold = veryBestScore - 500;

for (mc = oldList; mc != NULL; mc = next)
    {
    next = mc->next;
    if (mc->bestForRaft && mc->score >= threshold)
	{
	slAddHead(&newList, mc);
	}
    }
slReverse(&newList);
return newList;
}

struct raftFrag *findRaftFrag(struct raft *raft, struct oogFrag *frag)
/* Find frag in raft or die trying. */
{
struct raftFrag *rf;
for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    if (rf->frag == frag)
	return rf;
    }
errAbort("Couldn't find %s in raft", frag->name);
}

int scoreMrnaPsl(struct psl *psl)
/* Return score for one mRNA oriented psl. */
{
int milliBad;
int score;

milliBad = pslCalcMilliBad(psl, TRUE);
score = 25*log(1+psl->match) + log(1+psl->repMatch) - 10*milliBad;
if (psl->match <= 10)
    score -= (10-psl->match)*25;
score -= 10;
return score;
}


void findPslTails(struct psl *psl, int *retStartTail, int *retEndTail)
/* Find the length of "tails" (rather than extensions) implied by psl. */
{
int orientation = pslOrientation(psl);
int qFloppyStart, qFloppyEnd;
int tFloppyStart, tFloppyEnd;

if (orientation > 0)
    {
    qFloppyStart = psl->qStart;
    qFloppyEnd = psl->qSize - psl->qEnd;
    }
else
    {
    qFloppyStart = psl->qSize - psl->qEnd;
    qFloppyEnd = psl->qStart;
    }
tFloppyStart = psl->tStart;
tFloppyEnd = psl->tSize - psl->tEnd;
*retStartTail = min(qFloppyStart, tFloppyStart);
*retEndTail = min(qFloppyEnd, tFloppyEnd);
}

int scoreBacEndPsl(struct psl *psl)
/* Return score for one BAC end oriented psl. */
{
int milliBad;
int score;
int startTail, endTail;

milliBad = pslCalcMilliBad(psl, FALSE);
findPslTails(psl, &startTail, &endTail);
score = 25*log(1+psl->match) + log(1+psl->repMatch) - 20*milliBad - 4*(startTail+endTail);
if (psl->match <= 20)
    score -= (20-psl->match)*25;
return score;
}


struct raft *raftOfOne(struct oogFrag *frag)
/* Return a new raft containing just one fragment. */
{
struct raft *raft;
struct raftFrag *rf;

AllocVar(rf);
rf->orientation = 1;
rf->raftOffset = 0;
rf->frag = frag;

AllocVar(raft);
raft->id = nextRaftId();
raft->start = 0;
raft->end = frag->size;
slAddTail(&raft->fragList, rf);

frag->raft = raft;
return raft;
}

struct oogFrag *fragForName(struct hash *fragHash, char *name)
/* Return names fragment. */
{
struct oogFrag *frag;
struct hashEl *hel;

hel = hashLookup(fragHash, name);
if (hel == NULL)
    {
    errAbort("Conflict between self.psl and mrna.psl or est.psl, "
	     "not run on same data.  self.psl lacks %s", name);
    }
frag = hel->val;
return frag;
}

struct mrnaLine *oneRnaLine(struct psl *pslList, struct hash *fragHash)
/* Return an ordered list of rafts connected by this mRNA. */
{
struct mrnaLine *line;
struct mrnaConnection *mc, *mcList = NULL;
struct psl *psl;
struct oogFrag *frag;
struct raft *raft;
struct raftFrag *rf;
int oneScore;
int milliScore;
double conflictTotal;
int pslOrient;
int minScore = 0x3fffffff;
int totalScore = 0;
int averageScore;
int cleanSize = 0;
int compositeScore;
char *mrnaName = pslList->qName;

fprintf(logFile, "oneRnaLine (preCleaned):\n");
for (psl = pslList; psl != NULL; psl = psl->next)
    {
    pslOrient = pslOrientation(psl);
    AllocVar(mc);
    frag = fragForName(fragHash, psl->tName);
    raft = frag->raft;
    rf = findRaftFrag(raft, frag);
    mc->raft = raft;
    mc->orientation = pslOrient * rf->orientation;
    mc->score = scoreMrnaPsl(psl);
    fprintf(logFile, " %d %c %d %s\n", raft->id,
    	(mc->orientation > 0 ? '+' : '-'),
	mc->score, psl->tName);
    slAddHead(&mcList, mc);
    }
slReverse(&mcList);
fprintf(logFile, "%s got %d initial connections\n", mrnaName, slCount(mcList));
mcList = cleanMrnaConnectionList(mcList);
if (mcList == NULL)
    {
    fprintf(logFile, "Rejected %s\n", mrnaName);
    return NULL;
    }
if (mcList->next == NULL)
    {
    fprintf(logFile, "%s only hits one raft\n", mrnaName);
    return NULL;
    }
AllocVar(line);
line->name = cloneString(mrnaName);

/* Figure out composite score - average of average and min scores. */
for (mc = mcList; mc != NULL; mc = mc->next)
    {
    oneScore = mc->score;
    totalScore += oneScore;
    if (oneScore < minScore)
	minScore = oneScore;
    ++cleanSize;
    }
averageScore = (totalScore+cleanSize/2)/cleanSize;
compositeScore = (averageScore + minScore)/2;
fprintf(logFile, "   %d final connections score %d\n", 
    slCount(mcList), compositeScore);
line->score = compositeScore;
line->mcList = mcList;
return line;
}

boolean filterFrag(struct oogFrag *frag)
/* Return TRUE if passed filter. */
{
struct oogClone *clone = frag->clone;
if (clone->phase == htgFinished)
    return TRUE;
return frag->size >= minFragSize;
}

boolean pslTargetFragFilter(struct psl *psl, struct hash *fragHash)
/* See if target fragment passes filter. */
{
struct oogFrag *frag = fragForName(fragHash, psl->tName);
return filterFrag(frag);
}

boolean pslFragFilter(struct psl *psl, struct hash *fragHash)
/* See if both sides of psl pass filter. */
{
struct oogFrag *frag = fragForName(fragHash, psl->tName);
if (!filterFrag(frag)) return FALSE;
frag = fragForName(fragHash, psl->qName);
return filterFrag(frag);
}


struct mrnaLine *makeMrnaLines(char *fileName, struct hash *fragHash)
/* Return list of lines based on an mRNA psl file. */
{
struct hash *rnaHash = newHash(0);
struct mrnaLine *lineList = NULL, *line;
struct psl *pslList = NULL, *psl;
struct lineFile *lf = pslFileOpen(fileName);
struct hashEl *hel;
struct mrnaPsl *mpList = NULL, *mp;

/* Bundle together psl's that refer to same mRNA. */
while ((psl = pslNext(lf)) != NULL)
    {
    if (pslTargetFragFilter(psl, fragHash))
	{
	if ((hel = hashLookup(rnaHash, psl->qName)) == NULL)
	    {
	    AllocVar(mp);
	    hel = hashAdd(rnaHash, psl->qName, mp);
	    mp->mrnaName = hel->name;
	    slAddHead(&mpList, mp);
	    }
	else
	    mp = hel->val;
	slAddHead(&mp->pslList, psl);
	}
    }
slReverse(&mpList);

for (mp = mpList; mp != NULL; mp = mp->next)
    {
    slSort(&mp->pslList, pslCmpQuery);
    line = oneRnaLine(mp->pslList, fragHash);
    if (line != NULL)
	{
	slAddHead(&lineList, line);
	}
    }
slReverse(&lineList);
slFreeList(&mpList);
freeHash(&rnaHash);
return lineList;
}

void pairFree(struct fragPair **pPair)
/* Free a pair. */
{
struct fragPair *pair = *pPair;
pslFreeList(&pair->aPslList);
pslFreeList(&pair->bPslList);
freez(pPair);
}

void pairFreeList(struct fragPair **pList)
/* Free a list of pairs. */
{
struct fragPair *el, *next;
for (el = *pList; el != NULL; el = next)
    {
    next = el->next;
    pairFree(&el);
    }
*pList = NULL;
}

void makePairHash(char *pairName, struct hash **retHash, struct fragPair **retList)
/* Make up a hash out of the pair file. */
{
struct hash *hash = newHash(20);
struct hashEl *helA, *helB;
struct fragPair *list = NULL, *pair;
struct lineFile *lf = lineFileOpen(pairName, TRUE);
int lineSize;
char *line;
int wordCount;
char *words[3];
int pairCount = 0;

while (lineFileNext(lf, &line, &lineSize))
    {
    if ((wordCount = chopLine(line, words)) != 2)
	errAbort("Expecting 2 words line %d of %s\n", lf->lineIx, lf->fileName);
    AllocVar(pair);
    helA = hashAdd(hash, words[0], pair);
    helB = hashAdd(hash, words[1], pair);
    pair->a = helA->name;
    pair->b = helB->name;
    slAddHead(&list, pair);
    ++pairCount;
    }
printf("Got %d pairs in %s\n", pairCount, pairName);
lineFileClose(&lf);
slReverse(&list);
*retHash = hash;
*retList = list;
}

struct psl *bestBacPsl(struct psl *pslList, int *retScore)
/* Return best scoring psl in list according to bac scoring function,
 * or NULL if none of them are good enough. */
{
struct psl *psl;
int bestScore = -0x3fffffff;
int scoreOne;
struct psl *bestPsl = NULL;

for (psl = pslList; psl != NULL; psl = psl->next)
    {
    scoreOne = scoreBacEndPsl(psl);
    if (scoreOne > bestScore)
	{
	bestScore = scoreOne;
	bestPsl = psl;
	}
    }
*retScore = bestScore;
return bestPsl;
}

struct raft *raftFromBacPslList(struct psl *pslList, struct hash *fragHash, 
	int *retScore, struct psl **retPsl, struct raftFrag **retRf, struct oogFrag **retFrag)
/* Lookup raft from best scoring BAC psl in list. */
{
struct psl *psl;
struct oogFrag *frag;
struct hashEl *hel;
struct raft *raft;

if ((psl = bestBacPsl(pslList, retScore)) == NULL)
    return NULL;
if ((hel = hashLookup(fragHash, psl->tName)) == NULL)
    {
    errAbort("%s in bacEnds.psl but not in geno.lst files.  Need to regenerate .psl files?\n",
	psl->tName);
    }
*retPsl = psl;
*retFrag = frag = hel->val;
raft = frag->raft;
*retRf = findRaftFrag(raft, frag);
return raft;
}

struct oogFrag *fragInHash(struct hash *fragHash, char *fragName)
/* Find frag in hash. */
{
struct hashEl *hel = hashLookup(fragHash, fragName);
return hel->val;
}

struct mrnaLine *mlFromPair(struct fragPair *pair, struct hash *fragHash)
/* Create an mrnaLine describing connection between rafts implied by pair. */
{
struct raft *aRaft, *bRaft;
struct psl *aPsl, *bPsl;
struct raftFrag *aRf, *bRf;
int aScore, bScore;
int aOrient, bOrient;
struct hashEl *hel;
struct mrnaLine *ml;
struct mrnaConnection *mc;
char buf[256];
struct oogFrag *aFrag, *bFrag;
int cloneDif;

if ((aRaft = raftFromBacPslList(pair->aPslList, fragHash, &aScore, &aPsl, &aRf, &aFrag)) == NULL)
    return NULL;
if ((bRaft = raftFromBacPslList(pair->bPslList, fragHash, &bScore, &bPsl, &bRf, &bFrag)) == NULL)
    return NULL;
if (aRaft == bRaft)
    return NULL;

cloneDif = aFrag->clone->bargePos - bFrag->clone->bargePos;
if (cloneDif < 0) cloneDif = -cloneDif;

AllocVar(ml);
sprintf(buf, "%s_%s", pair->a, pair->b);
ml->name = cloneString(buf);
ml->score = min(aScore, bScore) - 50;
if (cloneDif > 200000)
    ml->score -= (cloneDif-200000)/1500;

AllocVar(mc);
mc->raft = aRaft;
mc->orientation = aOrient = pslOrientation(aPsl) * aRf->orientation;
mc->score = aScore;
slAddTail(&ml->mcList, mc);

AllocVar(mc);
mc->raft = bRaft;
mc->orientation = bOrient = -pslOrientation(bPsl) * bRf->orientation;
mc->score = bScore;
slAddTail(&ml->mcList, mc);

return ml;
}


struct mrnaLine *makeBacEndLines(char *bacPairName, char *bacPslName, struct hash *fragHash)
/* Read in pair file and psl file to create list of bac end connections
 * in mrnaLine format. */
{
struct mrnaLine *mlList = NULL, *ml;
struct hash *pairHash;
struct fragPair *pairList, *pair;
struct psl *psl;
struct lineFile *lf;
struct hashEl *hel;
char *bacName;
int totalPsl = 0;
int pairedPsl = 0;

makePairHash(bacPairName, &pairHash, &pairList);
lf = pslFileOpen(bacPslName);
while ((psl = pslNext(lf)) != NULL)
    {
    if (pslTargetFragFilter(psl, fragHash))
	{
	++totalPsl;
	bacName = psl->qName;
	if ((hel = hashLookup(pairHash, bacName)) == NULL)
	    {
	    pslFree(&psl);
	    continue;
	    }
	++pairedPsl;
	pair = hel->val;
	if (sameString(bacName, pair->a))
	    {
	    slAddHead(&pair->aPslList, psl);
	    }
	else
	    {
	    slAddHead(&pair->bPslList, psl);
	    }
	}
    }
fprintf(stdout, "%d paired bacEnd psls.  %d total bacEnd psls\n", pairedPsl, totalPsl);
fprintf(logFile, "%d paired bacEnd psls.  %d total bacEnd psls\n", pairedPsl, totalPsl);

for (pair = pairList; pair != NULL; pair = pair->next)
    {
    ml = mlFromPair(pair, fragHash);
    if (ml != NULL)
	slAddHead(&mlList, ml);
    }
pairFreeList(&pairList);
freeHash(&pairHash);
slReverse(&mlList);
return mlList;
}

struct oogClone *cloneFromName(char *fragName, struct oogClone **pList, struct hash *hash)
/* Return clone from hashtable if it exists already, otherwise make up new clone
 * store it on list and in hash, and return it. */
{
struct hashEl *hel;
char name[128];

fragToCloneName(fragName, name);
if ((hel = hashLookup(hash, name)) == NULL)
    errAbort("%s in self.psl but not geno.lst", name);
return hel->val;
}

struct oogFrag *fragFromName(char *name, struct oogFrag **pList, struct hash *hash,
	struct oogClone *clone, int size)
/* Return frag from hash table if it exists, otherwise make up a new one,
 * put it on list and hash table, and return it. */
{
struct hashEl *hel;

if ((hel = hashLookup(hash, name)) == NULL)
    {
    errAbort("%s is in self.psl, but not in any of the clones in %s\n"
	     "You probably need to regenerate self.psl");
    }
return hel->val;
}

boolean enclosingPsl(struct psl *psl)
/* Return TRUE if one or other fragment of PSL pretty much encloses
 * the other. */
{
if (psl->tStart <= 300 && psl->tEnd + 300 >= psl->tSize)
    return TRUE;
if (psl->qStart <= 300 && psl->qEnd + 300 >= psl->qSize)
    return TRUE;
return FALSE;
}

struct oogEdge *edgeFromSelfPsl(struct psl *psl,  struct oogFrag *qFrag, struct oogFrag *tFrag)
/* Return a self edge if this looks feasable, else NULL. */
{
bool qForward = (psl->strand[0] == '+');
int qFloppyStart, qFloppyEnd;
int tFloppyStart, tFloppyEnd;
struct oogEdge *edge;
int startTail, endTail;
int sizeFactor;

if (sameString(psl->qName, "AC007445.3_8") || sameString(psl->tName, "AC007445.3_8"))
{
findPslTails(psl, &startTail, &endTail); 
fprintf(logFile,
   "edgeFromSelf:\n"
   "  (%s %d %d of %d) %s (%s %d %d of %d) \n"
   "  startTail %d endTail %d\n",
    psl->qName, psl->qStart, psl->qEnd, psl->qSize, psl->strand,
    psl->tName, psl->tStart, psl->tEnd, psl->tSize,
    startTail, endTail);
}
#ifdef VERBOSE_LOG
#endif /* VERBOSE_LOG */

if (qForward)
    {
    qFloppyStart = psl->qStart;
    qFloppyEnd = psl->qSize - psl->qEnd;
    tFloppyStart = psl->tStart;
    tFloppyEnd = psl->tSize - psl->tEnd;
    if ((startTail = min(qFloppyStart,tFloppyStart)) > maxTailSize)
	return NULL;
    if ((endTail = min(qFloppyEnd, tFloppyEnd)) > maxTailSize)
	return NULL;
    AllocVar(edge);
    edge->orientation = 1;
    if (qFloppyStart >= tFloppyStart)
	{
	/*   >qqqqqq>        or   >qqqqqqqqqq>
	        >tttttt>             >tttt>
	 */
	edge->aSide = qFrag;
	edge->bSide = tFrag;
	edge->aCovers = (qFloppyEnd >= tFloppyEnd);
	}
    else
	{
	/*   >tttttt>        or   >tttttttttt>
	        >qqqqqq>             >qqqq>
	 */
	edge->aSide = tFrag;
	edge->bSide = qFrag;
	edge->aCovers = (tFloppyEnd >= qFloppyEnd);
	}
    }
else
    {
    qFloppyStart = psl->qSize - psl->qEnd;
    qFloppyEnd = psl->qStart;
    tFloppyStart = psl->tStart;
    tFloppyEnd = psl->tSize - psl->tEnd;
    if ((startTail = min(qFloppyStart,tFloppyStart)) > maxTailSize)
	return NULL;
    if ((endTail = min(qFloppyEnd, tFloppyEnd)) > maxTailSize)
	return NULL;
    AllocVar(edge);
    edge->orientation = -1;
    if (qFloppyStart >= tFloppyStart)
	{
	/*    <qqqqqqq<      or   <qqqqqqqqqqqqq<
	         >ttttttt>           >ttttt>
	 */
	edge->aSide = qFrag;
	edge->bSide = tFrag;
	edge->aCovers = (qFloppyEnd >= tFloppyEnd);
	}
    else
	{
	/*    >ttttttt>      or   >ttttttttttttt>
	         <qqqqqqq<           <qqqqq<
	 */
	edge->aSide = tFrag;
	edge->bSide = qFrag;
	edge->aCovers = (tFloppyEnd >= qFloppyEnd);
	}
    }
edge->match = psl->match;
edge->repMatch = psl->repMatch;
edge->milliBad = pslCalcMilliBad(psl, FALSE);
edge->tailA = startTail;
edge->tailB = endTail;
edge->score = -30*edge->milliBad - (startTail+endTail)/2 + 
	round(25*log(edge->match+1) + log(edge->repMatch+1)); 
/* Fold in penalties for not having any unique matches
 * or for being a small fragment. */
if (!enclosingPsl(psl))
    {
    if (edge->match <= 20)
	edge->score -= (20-edge->match)*25;
    if (psl->qSize < 5000)
	edge->score -= (5000-psl->qSize)/40;
    if (psl->tSize < 5000)
	edge->score -= (5000-psl->tSize)/40;
    }

edge->psl = psl;
if (sameString(psl->qName, "AC007445.3_8") || sameString(psl->tName, "AC007445.3_8"))
    fprintf(logFile, "  ~~~score %d milliBad %d minScore %d\n", edge->score,
	    edge->milliBad, minScore);
#ifdef VERBOSE_LOG
#endif /* VERBOSE_LOG */
return edge;
}

void processSelf(char *fileName, struct oogClone **pCloneList,
	struct oogFrag **pFragList, struct oogEdge **pEdgeList,
	struct psl **pSelfPslList,
	struct hash *cloneHash, struct hash *fragHash)
/* Process self.psl.  Make up clone, fragment, and self edge list. */
{
struct lineFile *lf = pslFileOpen(fileName);
struct psl *pslList = NULL, *psl;
struct oogEdge *edge;
struct oogClone *qClone, *tClone;
struct oogFrag *qFrag, *tFrag;

printf("processing %s\n", fileName);
while ((psl = pslNext(lf)) != NULL)
    {
    if (pslFragFilter(psl, fragHash))
	{
	qClone = cloneFromName(psl->qName, pCloneList, cloneHash);
	tClone = cloneFromName(psl->tName, pCloneList, cloneHash);
	qFrag = fragFromName(psl->qName, pFragList, fragHash, qClone, psl->qSize);
	tFrag = fragFromName(psl->tName, pFragList, fragHash, tClone, psl->tSize);
	if (!sameString(psl->qName, psl->tName))
	    {
	    edge = edgeFromSelfPsl(psl, qFrag, tFrag);
	    if (edge != NULL && edge->score >= minScore)
		{
		slAddHead(pEdgeList, edge);
		}
	    }
	slAddHead(&pslList, psl);
	}
    }
lineFileClose(&lf);
slReverse(&pslList);
slReverse(pEdgeList);
*pSelfPslList = pslList;
}


void fragRangeToRaftRange(int s, int e, int sz, short orientation, int raftOffset,
	int *retStart, int *retEnd)
/* Convert range from s to e inside fragment of size sz to raft coordinates. */
{
if (orientation > 0)
    {
    *retStart = raftOffset + s;
    *retEnd = raftOffset + e;
    }
else
    {
    *retStart = sz - e + raftOffset;
    *retEnd = sz - s + raftOffset;
    }
}

void fragAliRange(struct oogFrag *frag, struct psl *psl, int *retStart, int *retEnd)
/* Return range of aligning parts of frag. */
{
if (sameString(frag->name, psl->qName))
    {
    *retStart = psl->qStart;
    *retEnd = psl->qEnd;
    }
else
    {
    *retStart = psl->tStart;
    *retEnd = psl->tEnd;
    }
}

void pslFragToRaftAliRange(
	struct oogFrag *frag, short orientation, int raftOffset,
	struct psl *psl,
	int *retAliStart, int *retAliEnd)
/* Convert alignment range associated with psl and frag from frag
 * coordinates to raft coordinates, taking care of reversing if
 * on minus strand relative to raft. */
{
int s,e, sz;
fragAliRange(frag, psl, &s, &e);
sz = frag->size;
fragRangeToRaftRange(s, e, sz, orientation, raftOffset, retAliStart, retAliEnd);
}

void raftFragToRaftAliRange(struct raftFrag *rf, 
	struct psl *psl,
	int *retAliStart, int *retAliEnd)
/* Figure out start and end of alignment involving rf as defined by psl in
 * raft coordinates. */
{
pslFragToRaftAliRange(rf->frag, rf->orientation, rf->raftOffset,
	psl, retAliStart, retAliEnd);
}

void normalizeRaft(struct raft *raft)
/* Rearrange all offsets so that raft goes from zero to length of raft.  Fix up
 * graph begin and end.*/
{
int sMin = 0x3ffffff;
int eMax = -sMin;
struct raftFrag *rf;

for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    int s = rf->raftOffset;
    int e = s + rf->frag->size;
    if (s < sMin) sMin = s;
    if (e > eMax) eMax = e;
    }
raft->start = 0;
raft->end = eMax - sMin;
for (rf = raft->fragList; rf != NULL; rf = rf->next)
    rf->raftOffset -= sMin;
}



struct raft *raftOfTwo(struct oogEdge *edge)
/* Return a new raft containing just two fragments joined by edge. */
{
struct raft *raft;
struct raftFrag *rf;
struct psl *psl = edge->psl;
int aliSize;
int maxDiff;
int bOffset;
int orientation = edge->orientation;
int aAliStart, aAliEnd;
int bAliStart, bAliEnd;

AllocVar(raft);
raft->start = 0;
raft->id = nextRaftId();
fragAliRange(edge->aSide, psl, &aAliStart, &aAliEnd);
fragAliRange(edge->bSide, psl, &bAliStart, &bAliEnd);
if (orientation > 0)
    bOffset = aAliStart - bAliStart;
else
    {
    bOffset = aAliStart - (edge->bSide->size - bAliEnd);
    }
raft->end = max(edge->bSide->size + bOffset, edge->aSide->size);

/* Position first frag. */
AllocVar(rf);
rf->raftOffset = 0;
rf->orientation = 1;
rf->frag = edge->aSide;
edge->aSide->raft = raft;
slAddHead(&raft->fragList, rf);

/* Position second frag. */
AllocVar(rf);
rf->raftOffset = bOffset;
rf->orientation = orientation;
rf->frag = edge->bSide;
edge->bSide->raft = raft;
slAddHead(&raft->fragList, rf);

fprintf(logFile, "raftOfTwo: %s(%d) %c %s(%d) @ %d   .... psl (%d %d) (%d %d) start %d end %d\n", 
        edge->aSide->name,  edge->aSide->size,
	(edge->orientation > 0 ? '+' : '-'),
	edge->bSide->name, edge->bSide->size, bOffset, 
	psl->qStart, psl->qEnd, psl->tStart, psl->tEnd,
	raft->start, raft->end);
normalizeRaft(raft);
return raft;
}

/* Buffers to check raft/frag and raft/raft overlap. */
static int usedArrayAlloc = 0;
static bool *usedArray = NULL;

static void allocEnoughUsed(int checkSize)
/* Make sure usedArray is big enough for checkSize and
 * zero it out. */
{
if (checkSize > usedArrayAlloc)
    {
    usedArrayAlloc = checkSize + 10000;
    usedArray = needMoreMem(usedArray, 0, usedArrayAlloc);
    }
zeroBytes(usedArray, checkSize);
}

boolean fragMostlyAligns(struct oogFrag *frag, int raftOffset, short raftOrientation, 
	int fragStart, int fragEnd, struct raft *raft, struct oogEdge *selfEdgeList)
/* Return TRUE if frag aligns with something already in raft between start and stop. */
{
struct oogEdge *edge;
struct raftFrag *rf;
int raftStart;
int raftEnd;
int checkSize;
int misMatchCount = 0;
int i;

if (raftOrientation > 0)
    {
    raftStart = fragStart + raftOffset;
    raftEnd = fragEnd + raftOffset;
    }
else
    {
    int fSize = frag->size;
    raftStart = (fSize - fragEnd) + raftOffset;
    raftEnd = (fSize - fragStart) + raftOffset;
    }
if (raftStart < raft->start + maxTailSize)
    raftStart = raft->start + maxTailSize;
if (raftEnd > raft->end - maxTailSize)
    raftEnd = raft->end - maxTailSize;
checkSize = raftEnd - raftStart;

if (checkSize <= 0)
    return TRUE;

/* Allocate array that will track which parts have an alignment. */
allocEnoughUsed(checkSize);

for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    int offset = rf->raftOffset;
    char *rfName = rf->frag->name;
    bool relevant = FALSE;
    int s,e, sz;

    for (edge = selfEdgeList; edge != NULL; edge = edge->next)
	{
	struct psl *psl = edge->psl;
	int rStart, rEnd, rSize;

	if (raftOrientation*rf->orientation == edge->orientation)
	    {
	    if (sameString(psl->qName, frag->name) && sameString(psl->tName, rfName))
		{
		s = psl->tStart;
		e = psl->tEnd;
		sz = psl->tSize;
		relevant = TRUE;
		}
	    else if (sameString(psl->tName, frag->name) && sameString(psl->qName, rfName))
		{
		s = psl->qStart;
		e = psl->qEnd;
		sz = psl->qSize;
		relevant = TRUE;
		}
	    }
	if (relevant)
	    {
	    if (rf->orientation > 0)
		{
		rStart = s + offset;
		rEnd = e + offset;
		}
	    else
		{
		rStart = sz - e + offset;
		rEnd = sz - s + offset;
		}
	    if (rStart < raftStart) rStart = raftStart;
	    if (rEnd > raftEnd) rEnd = raftEnd;
	    rSize = rEnd - rStart;
	    if (rSize >= checkSize)
		return TRUE;   /* We're totally covered. */
	    if (rSize > 0)
		memset(usedArray + rStart - raftStart, 1, rSize);
	    }
	}
    }
for (i=0; i<checkSize; ++i)
    {
    if (!usedArray[i]) ++misMatchCount;
    }
return ( misMatchCount < 10);
}

void hashFragNames(struct hash *hash, struct raftFrag *rfList)
/* Put names of all frags in list onto hash. */
{
struct raftFrag *rf;
for (rf = rfList; rf != NULL; rf = rf->next)
    {
    hashAdd(hash, rf->frag->name, rf);
    }
}

boolean bMostlyAlignsWithA(struct raft *aRaft, int aStart, int aEnd, struct raft *bRaft,
   int bStart, int bEnd, short relOrientation, struct oogEdge *mergeEdge, 
   struct oogEdge *selfEdgeList, int diag)
/* Return TRUE if the indicated part of b aligns roughly with the indicated part
 * of a. */
{
struct hash *aHash = newHash(0), *bHash = newHash(0);
int checkSize = aEnd - aStart;
int bCheckSize = bEnd - bStart;
struct oogEdge *edge;
int i;
int misMatchCount = 0;
int aOk = FALSE;


if (checkSize <= maxTailSize)
    return TRUE;
if (bCheckSize != checkSize)
    {
    dumpEdge(mergeEdge, logFile);
    errAbort("internal error: aCheckSize(%d) != bCheckSize(%d)", checkSize, bCheckSize);
    }
if (aStart < aRaft->start)
    {
    dumpEdge(mergeEdge, logFile);
    errAbort("internal error: aStart(%d) < aRaft->start(%d)", aStart, aRaft->start);
    }
if (bStart < bRaft->start)
    {
    dumpEdge(mergeEdge, logFile);
    errAbort("internal error: bStart(%d) < bRaft->start(%d)", bStart, bRaft->start);
    }
if (aEnd > aRaft->end)
    {
    dumpEdge(mergeEdge, logFile);
    errAbort("internal error: aEnd(%d) > aRaft->end(%d)", aEnd, aRaft->end);
    }
if (bEnd > bRaft->end)
    {
    dumpEdge(mergeEdge, logFile);
    errAbort("internal error: bEnd(%d) > bRaft->end(%d)", bEnd, bRaft->end);
    }


hashFragNames(aHash, aRaft->fragList);
hashFragNames(bHash, bRaft->fragList);

allocEnoughUsed(checkSize);
for (edge = selfEdgeList; edge != NULL; edge = edge->next)
    {
    struct psl *psl = edge->psl;
    struct hashEl *aHel, *bHel;
    int as,ae,bs,be,asz,bsz;
    boolean relevant = FALSE;

    if ((aHel = hashLookup(aHash, psl->qName)) != NULL 
    	&& ((bHel = hashLookup(bHash, psl->tName)) != NULL))
	{
	as = psl->qStart;
	ae = psl->qEnd;
	asz = psl->qSize;
	bs = psl->tStart;
	be = psl->tEnd;
	bsz = psl->tSize;
	relevant = TRUE;
	}
    else if ((aHel = hashLookup(aHash, psl->tName)) != NULL
	&& ((bHel = hashLookup(bHash, psl->qName)) != NULL))
	{
	as = psl->tStart;
	ae = psl->tEnd;
	asz = psl->tSize;
	bs = psl->qStart;
	be = psl->qEnd;
	bsz = psl->qSize;
	relevant = TRUE;
	}
    if (relevant)
	{
	struct raftFrag *aRf = aHel->val, *bRf = bHel->val;
	int lDiag;
	int diagDistance;

	fragRangeToRaftRange(as, ae, asz, aRf->orientation, aRf->raftOffset, &as, &ae);
	fragRangeToRaftRange(bs, be, bsz, bRf->orientation, bRf->raftOffset, &bs, &be);
	if (relOrientation > 0)
	    lDiag = as - bs;
	else
	    lDiag = as - (bRaft->end - be);
	diagDistance = intAbs(lDiag - diag);
	if (diagDistance < 10)
	    {
	    int s,e,sz;
	    s = as - aStart;
	    if (s < 0)
		s = 0;
	    e = ae - aStart;
	    if (e > checkSize)
		e = checkSize;
	    sz = e - s;
	    fprintf(logFile, "as %d, ae %d, s %d, e %d, sz %d, checkSize %d\n",
	    	as, ae, s, e, sz, checkSize);
	    if (sz == checkSize)
		{
		aOk = TRUE;	/* We're totally covered... */
		break;
		}
	    if (sz > 0)
		{
		memset(usedArray + s, 1, sz);
		}
	    }
	}
    }
if (!aOk)
    {
    for (i=0; i<checkSize; ++i)
	{
	if (!usedArray[i]) ++misMatchCount;
	}
    }

freeHash(&aHash);
freeHash(&bHash);
return (misMatchCount < 16);
}



void addEdgeRefToList(struct oogEdge *edge, struct edgeRef **pList)
/* Add reference to edge to raft. */
{
struct edgeRef *er;
AllocVar(er);
er->edge = edge;
slAddHead(pList, er);
}

void addEdgeRefToRaft(struct raft *raft, struct oogEdge *edge)
/* Add reference to edge to raft. */
{
addEdgeRefToList(edge, &raft->refList);
}

boolean canMerge(struct oogFrag *frag, struct oogEdge *staple, struct raftFrag *socket, 
	struct raft *raft, struct oogEdge *selfEdgeList, int *retOffset, short *retOrientation)
/* Return TRUE if frag can be merged into raft  Staple is the edge wanting
 * to bring frag into raft. Socket is the raftFrag corresponding to the
 * other frag in staple, which is already part of raft. */
{
int offset;
short orientation;
int raftAliStart, raftAliEnd;
int fragAliStart, fragAliEnd;
struct psl *psl = staple->psl;
int ok;

*retOrientation = orientation = socket->orientation * staple->orientation;
raftFragToRaftAliRange(socket, psl, &raftAliStart, &raftAliEnd);
fragAliRange(frag, psl, &fragAliStart, &fragAliEnd);
if (orientation > 0)
    *retOffset = offset = raftAliStart - fragAliStart;
else
    *retOffset = offset = raftAliStart - (frag->size - fragAliEnd);
ok = fragMostlyAligns(frag, offset, orientation, maxTailSize, frag->size-maxTailSize, raft, selfEdgeList);
return ok;
}

void noteRedundant(struct oogEdge *edge)
/* Note that given edge is redundant. */
{
fprintf(logFile, "redundant edge from %s to %s\n", edge->aSide->name, edge->bSide->name);
}

void noteConflicting(struct oogEdge *edge, char *why)
/* Note that given edge is redundant. */
{
fprintf(logFile, "conflicting edge from %s to %s (%s).\n", edge->aSide->name, edge->bSide->name, why);
}

boolean extendRaft(struct raft *raft, struct raftFrag *socket, 
	struct oogEdge *edge, struct oogFrag *newFrag, struct oogEdge *selfEdgeList)
/* Try to extend raft to include other fragment in edge. */
{
int offset;
short orientation;

fprintf(logFile, "extend raft\n");
if (canMerge(newFrag, edge, socket, raft, selfEdgeList, &offset, &orientation))
    {
    struct raftFrag *rf;
    int rfragStart, rfragEnd;
    AllocVar(rf);
    rf->raftOffset = offset;
    rf->frag = newFrag;
    rf->orientation = orientation;
    slAddHead(&raft->fragList, rf);
    newFrag->raft = raft;
    rfragStart = offset;
    rfragEnd = offset + newFrag->size;
    if (rfragStart < raft->start)
	raft->start = rfragStart;
    if (rfragEnd > raft->end)
	raft->end = rfragEnd;
    fprintf(logFile, "extends ok @%d %c to %d %d\n", offset, (orientation > 0 ? '+' : '-'),
    	raft->start, raft->end);
    normalizeRaft(raft);
    return TRUE;
    }
else
    {
    noteConflicting(edge, "Can't extend raft");
    return FALSE;
    }
}

void flipRaftFrags(struct raft *raft)
/* Flip orientation of each frag within raft. */
{
int end = raft->end;
struct raftFrag *rf;

fprintf(logFile, "Flipping raft %d\n", raft->id);
for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    rf->orientation = -rf->orientation;
    rf->raftOffset = end - (rf->raftOffset + rf->frag->size);
    }
raft->end = end - raft->start;
raft->start = 0;
slReverse(&raft->fragList);
}

void swallowRaftB(struct raft *aRaft, struct raft *bRaft, 
	int aAliStart, int bAliStart, int aAliEnd, int bAliEnd, int orientation)
/* Make raft A swallow raft B. */
{
struct raftFrag *rf;
int bStart, bEnd;
int offset;

if (orientation < 0)
    {
    /* First flip around raft B. */
    int oldAliStart;
    int bEnd = bRaft->end;

    flipRaftFrags(bRaft);

    /* Make corresponding flip in b alignment coordinates. */
    oldAliStart = bAliStart;
    bAliStart = bEnd - bAliEnd;
    bAliEnd = bEnd - oldAliStart;
    }

offset = aAliStart - bAliStart;
for (rf = bRaft->fragList; rf != NULL; rf = rf->next)
    {
    rf->raftOffset += offset;
    rf->frag->raft = aRaft;
    }
bStart = bRaft->start + offset;
bEnd = bRaft->end + offset;
if (bStart < aRaft->start)
    aRaft->start = bStart;
if (bEnd > aRaft->end)
    aRaft->end = bEnd;

aRaft->fragList = slCat(aRaft->fragList, bRaft->fragList);
bRaft->fragList = NULL;
aRaft->refList = slCat(aRaft->refList, bRaft->refList);
bRaft->refList = NULL;

fprintf(logFile, "Merged rafts at offset %d orient %d start %d end %d\n", offset, orientation,
    aRaft->start, aRaft->end);
normalizeRaft(aRaft);
}

boolean canMergeRaftInMiddle(struct oogEdge *selfEdgeList, struct raft *aRaft, struct raft *bRaft,
	struct raftFrag *aRf, struct raftFrag *bRf, struct oogEdge *edge,
	int relativeRaftOrientation, int aRaftAliStart, int aRaftAliEnd,
	int bRaftAliStart, int bRaftAliEnd, int startTail, int endTail)
/* Here we have an edge that connects the middle of two rafts.  We need to
 * make sure that the rafts are compatable along the full length of their
 * overlap, not just along the part aligned by edge. */
{
int as, bs, diag;
fprintf(logFile, "canMergeRaftInMiddle(raftA %d:%d-%d of %d, raftB %d:%d-%d of %d, relOrient %d)\n",
	aRaft->id, aRaftAliStart, aRaftAliEnd, aRaft->end,
	bRaft->id, bRaftAliStart, bRaftAliEnd, bRaft->end,
	relativeRaftOrientation);
if (relativeRaftOrientation > 0)
    {
    fprintf(logFile, " >>>case 1<<<\n");
    as = aRaftAliStart;
    bs = bRaftAliStart;
    diag = as-bs;
    if (!bMostlyAlignsWithA(aRaft, aRaftAliStart-startTail, aRaftAliStart, 
    	bRaft, bRaftAliStart - startTail, bRaftAliStart, 
    	relativeRaftOrientation, edge, selfEdgeList, diag))
	return FALSE;
    if (!bMostlyAlignsWithA(aRaft, aRaftAliEnd, aRaftAliEnd+endTail, 
    	bRaft, bRaftAliEnd, bRaftAliEnd+endTail,
    	relativeRaftOrientation, edge, selfEdgeList, diag))
	return FALSE;
    }
else
    {
    fprintf(logFile, " >>>case -1<<<\n");
    as = aRaftAliStart;
    bs = bRaft->end - bRaftAliEnd;
    diag = as-bs;
    if (!bMostlyAlignsWithA(aRaft, aRaftAliStart-startTail, aRaftAliStart, 
    	bRaft, bRaftAliEnd, bRaftAliEnd+startTail, 
    	relativeRaftOrientation, edge, selfEdgeList, diag))
	return FALSE;
    if (!bMostlyAlignsWithA(aRaft, aRaftAliEnd, aRaftAliEnd+endTail, 
    	bRaft, bRaftAliStart-endTail, bRaftAliStart, 
    	relativeRaftOrientation, edge, selfEdgeList, diag))
	return FALSE;
    }
return TRUE;
}


boolean mergeRafts(struct raft **pRaftList, struct oogEdge *selfEdgeList, 
	struct raft *aRaft, struct raft *bRaft, 
	struct raftFrag *aRf, struct raftFrag *bRf, struct oogEdge *edge)
/* Try and merge two rafts joined by edge. */
{
int aRaftAliStart, aRaftAliEnd, bRaftAliStart, bRaftAliEnd;
int startTail, endTail;
struct psl *psl = edge->psl;
int relativeRaftOrientation = edge->orientation * aRf->orientation * bRf->orientation;

fprintf(logFile, "mergeRafts %d %d\n", aRaft->id, bRaft->id);

raftFragToRaftAliRange(aRf, psl, &aRaftAliStart, &aRaftAliEnd);
raftFragToRaftAliRange(bRf, psl, &bRaftAliStart, &bRaftAliEnd);

fprintf(logFile, "  a %s pos %d orient %d size %d raft %d raftStart %d raftEnd %d\n",
    aRf->frag->name, aRf->raftOffset, aRf->orientation, aRf->frag->size,
    aRaft->id, aRaft->start, aRaft->end);
fprintf(logFile, "  b %s pos %d orient %d size %d raft %d raftStart %d raftEnd %d\n",
    bRf->frag->name, bRf->raftOffset, bRf->orientation, bRf->frag->size,
    bRaft->id, bRaft->start, bRaft->end);


if (relativeRaftOrientation > 0)
    {
    startTail = min(aRaftAliStart - aRaft->start, bRaftAliStart - bRaft->start);
    endTail = min(aRaft->end - aRaftAliEnd, bRaft->end - bRaftAliEnd);
    }
else
    {
    startTail = min(aRaftAliStart - aRaft->start, bRaft->end - bRaftAliEnd);
    endTail = min(aRaft->end - aRaftAliEnd, bRaftAliStart - bRaft->start);
    }

if (startTail > maxTailSize)
    {
    noteConflicting(edge, "Merge long start tail");
    fprintf(logFile, " startTail %d endTail %d\n", startTail, endTail);
    if (!canMergeRaftInMiddle(selfEdgeList, aRaft, bRaft, aRf, bRf, edge, relativeRaftOrientation,
    	aRaftAliStart, aRaftAliEnd, bRaftAliStart, bRaftAliEnd, startTail, endTail))
	return FALSE;
    }
else if (endTail > maxTailSize)
    {
    noteConflicting(edge, "Merge long end tail");
    fprintf(logFile, " startTail %d endTail %d\n", startTail, endTail);
#ifdef FLAKY
    if (!canMergeRaftInMiddle(selfEdgeList, aRaft, bRaft, aRf, bRf, edge, relativeRaftOrientation,
    	aRaftAliStart, aRaftAliEnd, bRaftAliStart, bRaftAliEnd, startTail, endTail))
#endif
	return FALSE;
    }
swallowRaftB(aRaft, bRaft, aRaftAliStart, bRaftAliStart, aRaftAliEnd, bRaftAliEnd, relativeRaftOrientation);
slRemoveEl(pRaftList, bRaft);
bRaft->parent = aRaft;
slAddHead(&aRaft->children, bRaft);
return TRUE;
}

void maybeLogRaft(struct raft *raft)
/* Dump raft to log file if verbose. */
{
#ifdef VERBOSE_LOG
dumpRaft(raft, logFile);
#endif
}

struct mergeFailure
/* Keep track of a failure to merge raft.  Later on the
 * rafts might have been merged for other reasons, in which
 * case we want to add the failed edge back to the raft to
 * give the golden path maker more to work with. */
    {
    struct mergeFailure *next;	/* Next in list. */
    struct oogEdge *edge;       /* Edge that failed to merge. */
    struct raft *aRaft, *bRaft; /* Two rafts to merge. */
    struct raftFrag *aRf, *bRf; /* Two frags to mergs. */
    };

struct raft *raftAncientAncestor(struct raft *raft)
/* Return eldest ancestor in raft tree. */
{
while (raft->parent != NULL)
    raft = raft->parent;
return raft;
}

void reprocessMergeFailures(struct mergeFailure *mergeFailureList)
/* Try adding back edged in mergeFailure list to rafts. */
{
struct mergeFailure *mf;
struct raft *parent;

for (mf = mergeFailureList; mf != NULL; mf = mf->next)
    {
    parent = raftAncientAncestor(mf->aRaft);
    if (parent == raftAncientAncestor(mf->bRaft))
	{
	fprintf(logFile, "Upon further consideration the following edge merges ok:\n");
	dumpEdge(mf->edge, logFile);
	addEdgeRefToRaft(parent, mf->edge);
	}
    }
}


struct oogEdge *findBestRaftEdge(struct oogFrag *frag, struct raft *origRaft,
	struct oogEdge *origEdge, struct oogEdge *edgeList, 
	struct raft *raftList)
/* Find biggest raft that frag could glom into reasonably.  */
{
int scoreThreshold;
struct raft *longestRaft = origRaft;
int longestSize = origRaft->end - origRaft->start;
struct oogEdge *bestEdge = origEdge;
struct oogEdge *edge;
struct raft *raft;
int size;

scoreThreshold = origEdge->score - 100;
if (scoreThreshold < minScore)
    scoreThreshold = minScore;
for (edge = edgeList; edge != NULL; edge = edge->next)
    {
    if (edge->score >= scoreThreshold && !edge->used)
	{
	if (edge->aSide == frag)
	    {
	    if ((raft = edge->bSide->raft) != NULL)
		{
		size = raft->end - raft->start;
		if (size > longestSize)
		    {
		    longestSize = size;
		    bestEdge = edge;
		    longestRaft = raft;
		    }
		}
	    }
	else if (edge->bSide == frag)
	    {
	    if ((raft = edge->aSide->raft) != NULL)
		{
		size = raft->end - raft->start;
		if (size > longestSize)
		    {
		    longestSize = size;
		    bestEdge = edge;
		    longestRaft = raft;
		    }
		}
	    }
	}
    }
if (origRaft != longestRaft)
    {
    logg("%s hits raft %d by %d score %d, better than raft %d by %d score %d\n",
	frag->name, longestRaft->id, longestSize, bestEdge->score,
	origRaft->id, (origRaft->end - origRaft->start), origEdge->score);
    }
return bestEdge;
}

void mergeBestRaftEdge(struct oogFrag *frag, struct raft *origRaft,
	struct oogEdge *origEdge, struct oogEdge *edgeList,
	struct raft *raftList, struct oogEdge *selfEdgeList)
/* Merge into best reasonable looking raft. */
{
struct oogEdge *e = findBestRaftEdge(frag, origRaft, origEdge, edgeList, raftList);
struct oogFrag *otherFrag;
struct raft *raft;
boolean edgeOk;

e->used = TRUE;
if (e->aSide == frag)
    {
    otherFrag = e->bSide;
    }
else if (e->bSide == frag)
    {
    otherFrag = e->aSide;
    }
else
    {
    assert(FALSE);
    }
raft = otherFrag->raft;
assert(raft != NULL);
edgeOk = extendRaft(raft, findRaftFrag(raft, otherFrag), e, frag, selfEdgeList);
if (edgeOk)
    {
    addEdgeRefToRaft(raft, e);
    maybeLogRaft(raft);
    }
}

struct raft *makeRafts(struct oogFrag *fragList, struct oogEdge *selfEdgeList)
/* Make connected rafts of fragments. Returns doubly linked list of rafts. */
{
struct raft *raftList = NULL;
struct oogEdge *edge;
struct raft *aRaft, *bRaft, *raft;
struct raftFrag *aRf, *bRf, *rf;
boolean aIn, bIn;
int max = 10;
struct mergeFailure *mergeFailureList = NULL;

for (edge = selfEdgeList; edge != NULL; edge = edge->next)
    edge->used = FALSE;
for (edge = selfEdgeList; edge != NULL; edge = edge->next)
    {
    struct psl *psl = edge->psl;
    boolean edgeOk = TRUE;

    dumpEdge(edge, logFile);
    fprintf(logFile, "psl: (%s %d %d of %d) %s (%s %d %d of %d)\n", 
	   psl->qName, psl->qStart, psl->qEnd, psl->qSize, 
	   psl->strand,
	   psl->tName, psl->tStart, psl->tEnd, psl->tSize);
    aIn = fragInRaft(raftList, edge->aSide, &aRaft, &aRf);
    bIn = fragInRaft(raftList, edge->bSide, &bRaft, &bRf);
    if (!aIn && !bIn)
	{
	raft = raftOfTwo(edge);
	addEdgeRefToRaft(raft, edge);
	maybeLogRaft(raft);
	slAddHead(&raftList, raft);
	}
    else if (aIn && bIn)
	{
	if (aRaft != bRaft)
	    {
	    edgeOk = mergeRafts(&raftList, selfEdgeList, aRaft, bRaft, aRf, bRf, edge);
	    if (edgeOk)
		{
		addEdgeRefToRaft(aRaft, edge);
		maybeLogRaft(aRaft);
		}
	    else
		{
		/* Need to look at this edge again later... */
		struct mergeFailure *mf;
		AllocVar(mf);
		mf->edge = edge;
		mf->aRaft = aRaft;
		mf->bRaft = bRaft;
		mf->aRf = aRf;
		mf->bRf = bRf;
		slAddHead(&mergeFailureList, mf);
		}
	    }
	else
	    {
	    if (aRf->orientation*bRf->orientation == edge->orientation)
		{
		noteRedundant(edge);
		addEdgeRefToRaft(aRaft, edge);
		maybeLogRaft(aRaft);
		}
	    else
		{
		edgeOk = FALSE;
		fprintf(logFile, "bad orientation in redundant edge\n");
		}
	    }
	}
    else if (aIn)
	{
	mergeBestRaftEdge(edge->bSide, aRaft, edge, edge->next, raftList, selfEdgeList);
	}
    else if (bIn)
	{
	mergeBestRaftEdge(edge->aSide, bRaft, edge, edge->next, raftList, selfEdgeList);
	}
    fprintf(logFile, "\n");
    if (edgeOk && raftPsl != NULL)
	pslTabOut(psl, raftPsl);
    }
reprocessMergeFailures(mergeFailureList);
slFreeList(&mergeFailureList);

for (raft = raftList; raft != NULL; raft = raft->next)
    {
    slSort(&raft->fragList, raftFragCmpOffset);
    }
slReverse(&raftList);
return raftList;
}

int raftedFragCount(struct oogFrag *fragList)
/* Return number of fragments that are part of a raft. */
{
struct oogFrag *frag;
int count = 0;
for (frag = fragList; frag != NULL; frag = frag->next)
    if (frag->raft != NULL)
	++count;
return count;
}

void makeSingletonRafts(struct oogFrag *fragList, struct raft **pRaftList)
/* Make up rafts around isolated fragments. */
{
struct oogFrag *frag;
struct raft *list = NULL, *raft;

for (frag = fragList; frag != NULL; frag = frag->next)
    {
    if (frag->raft == NULL && filterFrag(frag))
	{
	raft = raftOfOne(frag);
	slAddHead(&list, raft);
	}
    }
slReverse(&list);
*pRaftList = slCat(*pRaftList, list);
}

static struct oogClone *ovlpA, *ovlpB;	/* Clones being checked for overlaps. */
static bool *ovlpHits;                  /* A bool for each base in ovlpA, TRUE if overlaps. */
static int ovlpAlloc;                   /* Size allocated for ovlpHits. */

static void startCloneOverlapEstimate(struct oogClone *a, struct oogClone *b)
/* Start things up to estimate overlaps between clones a and b */
{
int aSize;
ovlpA = a;
ovlpB = b;
aSize = a->size;
if (a->size > ovlpAlloc)
    {
    ovlpAlloc = a->size+100000;
    ExpandArray(ovlpHits, 0, ovlpAlloc);
    }
memset(ovlpHits, 0, aSize);
}

#ifdef OLD
static void addEdgesToCloneOverlap(struct oogClone *a, struct oogClone *b, struct edgeRef *refList)
/* Add in edges to clone overlap. */
{
struct edgeRef *ref;
struct oogEdge *edge;
struct psl *psl;
struct oogFrag *aFrag;
int fragStart;
int fragEnd;
int fragOffset;
int size;
int aStart;

assert(a == ovlpA && b == ovlpB);
for (ref = refList; ref != NULL; ref = ref->next)
    {
    edge = ref->edge;
    if (edge->aSide->clone == a && edge->bSide->clone == b)
	aFrag = edge->aSide;
    else if (edge->bSide->clone == a && edge->aSide->clone == b)
	aFrag = edge->bSide;
    else
    	continue;
    psl = edge->psl;
    if (sameString(psl->qName, aFrag->name))
	{
	fragStart = psl->qStart;
	fragEnd = psl->qEnd;
	}
    else if (sameString(psl->tName, aFrag->name))
	{
	fragStart = psl->tStart;
	fragEnd = psl->tEnd;
	}
    else
	{
	assert(FALSE);
	}
    fragOffset = aFrag->clonePos;
    size = fragEnd - fragStart;
    aStart = fragOffset + fragStart;
    if (aStart + size > a->size)
	{
	size = a->size - aStart;
	if (size <= 0)
	    size = 0;
	}
    memset(ovlpHits+aStart, TRUE, size);
    }
}
#endif /* OLD */

static void addPslToCloneOverlap(struct oogClone *a, struct oogClone *b, 
    struct oogFrag *aFrag, struct hash *fragHash,
    struct psl *psl)
/* Add in edges to clone overlap. */
{
int fragStart;
int fragEnd;
int fragOffset;
int size;
int aStart;

assert(a == ovlpA && b == ovlpB);
if (sameString(psl->qName, aFrag->name))
    {
    fragStart = psl->qStart;
    fragEnd = psl->qEnd;
    }
else if (sameString(psl->tName, aFrag->name))
    {
    fragStart = psl->tStart;
    fragEnd = psl->tEnd;
    }
else
    {
    assert(FALSE);
    }
fragOffset = aFrag->clonePos;
size = fragEnd - fragStart;
aStart = fragOffset + fragStart;
if (aStart + size > a->size)
    {
    size = a->size - aStart;
    if (size <= 0)
	size = 0;
    }
memset(ovlpHits+aStart, TRUE, size);
}

int tallyCloneOverlap()
/* Tally total overlap between clones. */
{
int i;
int total = 0;
bool *hits = ovlpHits;
int size = ovlpA->size;

for (i=0; i<size; ++i)
    total += hits[i];
return total;
}

boolean cloneInRaft(struct oogClone *clone, struct raft *raft)
/* Return TRUE if clone is in raft. */
{
struct raftFrag *rf;
for (rf = raft->fragList; rf != NULL; rf = rf->next)
    if (rf->frag->clone == clone)
	return TRUE;
return FALSE;
}

boolean pslGoodEnough(struct psl *psl, struct aliFilter *filter)
/* Check if psl looks good enough according to given parameters. */
{
int startTail, endTail;

if (psl->qSize < filter->minFragSize || psl->tSize < filter->minFragSize)
    return FALSE;
if (psl->match < filter->minUnique)
    return FALSE;
if (psl->match + psl->repMatch < filter->minMatch)
    return FALSE;
findPslTails(psl, &startTail, &endTail);
if (startTail > filter->maxTail || endTail > filter->maxTail)
    return FALSE;
if (pslCalcMilliBad(psl, FALSE) > filter->maxBad)
    return FALSE;
return TRUE;
}


int findOverlapBetweenClones(struct oogClone *aClone, 
	struct oogClone *bClone, struct aliFilter *filter, 
	struct hash *fragHash, struct hash *cloneHash, struct psl *pslList)
/* Return approximate number of base pairs that overlap between
 * clones a and b. */
{
struct psl *psl;
bool possibleOverlap = FALSE;
struct oogFrag *aFrag, *frag;

startCloneOverlapEstimate(aClone, bClone);
for (psl = pslList; psl != NULL; psl = psl->next)
    {
    /* See if psl is between aClone and bClone. */
    frag = hashMustFindVal(fragHash, psl->qName);
    if (frag->clone == aClone)
        {
	aFrag = frag;
	frag = hashMustFindVal(fragHash, psl->tName);
	if (frag->clone != bClone)
	    continue;
	}
    else if (frag->clone == bClone)
        {
	frag = hashMustFindVal(fragHash, psl->tName);
	if (frag->clone != aClone)
	    continue;
	aFrag = frag;
	}
    else
        continue;    

    /* Add to overlap if it quality is good enough. */	
    if (pslGoodEnough(psl, filter))
	{
	possibleOverlap = TRUE;
	addPslToCloneOverlap(aClone, bClone, aFrag, fragHash, psl);
	}
    }
if (!possibleOverlap)
    return 0;
return tallyCloneOverlap();
}

char *ocpHashName(char *a, char *b)
/* Return name of ocp associated with a and b */
{
static char buf[256];
if (strcmp(a,b) > 0)
    sprintf(buf, "%s^%s", a, b);
else
    sprintf(buf, "%s^%s", b, a);
return buf;
}

struct clonePair *findOverlappingClones(struct oogClone *cloneList, struct aliFilter *filter, 
	struct psl *pslList, struct hash *fragHash, struct hash *cloneHash, struct hash *ocpHash)
/* Return list of overlapping clone pairs. */
{
struct clonePair *ocpList = NULL, *ocp;
struct oogClone *a, *b;
int overlap;
char *ocpName;

for (a = cloneList; a != NULL; a = a->next)
    {
    for (b = a->next; b != NULL; b = b->next)
	{
	overlap = findOverlapBetweenClones(a, b, filter, fragHash, cloneHash, pslList);
	if (overlap > 0)
	    {
	    AllocVar(ocp);
	    ocp->a = a;
	    ocp->b = b;
	    ocp->overlap = overlap;
	    slAddHead(&ocpList, ocp);
	    ocpName = ocpHashName(a->name, b->name);
	    hashAdd(ocpHash, ocpName, ocp);
	    }
	}
    }
slReverse(&ocpList);
fprintf(logFile, "Found %d overlapping pairs in %d clones\n", slCount(ocpList), slCount(cloneList));
return ocpList;
}


void saveOcp(char *fileName, struct clonePair *ocpList)
/* Save ocpList to file. */
{
FILE *f = mustOpen(fileName, "w");
struct clonePair *ocp;
for (ocp = ocpList; ocp != NULL; ocp = ocp->next)
    fprintf(f, "%-8s\t%-8s\t%d\n", ocp->a->name, ocp->b->name, ocp->overlap);
fclose(f);
}

int findHashedOverlap(char *aName, char *bName, struct hash *ocpHash)
/* Find how much a and b overlap. (obsolete?) */
{
struct hashEl *hel;
struct clonePair *ocp;
char *ocpName;

ocpName = ocpHashName(aName, bName);
if ((hel = hashLookup(ocpHash, ocpName)) == NULL)
    {
    ocpName = ocpHashName(bName, aName);
    hel = hashLookup(ocpHash, ocpName);
    }
if (hel == NULL)
    return 0;
ocp = hel->val;
return ocp->overlap;
}

int lookupOverlap(struct hash *hash, struct oogClone *a, struct oogClone *b)
/* Find overlap between two clones. */
{
struct clonePair *pair;
boolean firstA;

pair = hashFindVal(hash, ocpHashName(a->name, b->name));
if (pair == NULL)
    return 0;
return pair->overlap;
}

void sharesFromPairs(struct clonePair *pairList)
/* Make sequence connections between clones. */
{
struct clonePair *pair;
struct seqSharer *share;
struct oogClone *a, *b;
int overlap;

for (pair = pairList; pair != NULL; pair = pair->next)
    {
    overlap = pair->overlap;
    a = pair->a;
    b = pair->b;

    AllocVar(share);
    share->clone = a;
    share->pair = pair;
    slAddHead(&b->sharers, share);

    AllocVar(share);
    share->clone = b;
    share->pair = pair;
    slAddHead(&a->sharers, share);
    }
}

void findMaxOverlaps(struct oogClone *cloneList, struct hash *looseHash)
/* Find maximum overlapping clones from cloneList */
{
struct oogClone *a, *b;
struct seqSharer *share;
int maxOverlap;
int overlap;

for (a = cloneList; a != NULL; a = a->next)
    {
    maxOverlap = 0;
    for (share = a->sharers; share != NULL; share = share->next)
	{
	b = share->clone;
	overlap = lookupOverlap(looseHash, a, b);
	if (overlap > maxOverlap)
	    {
	    maxOverlap = overlap;
	    a->maxOverlapClone = b;
	    }
	}
    if (a->maxOverlapClone != NULL)
	a->maxOverlapPpt = round(1000.0*maxOverlap/a->size);
    }
}

void countUses(struct clonePair *pairList)
/* Fill in aUses and bUses fields of pairs. */
{
struct clonePair *pair;
struct oogClone *a, *b;

for (pair = pairList; pair != NULL; pair = pair->next)
    {
    a = pair->a;
    b = pair->b;
    pair->aUses = slCount(a->sharers);
    pair->bUses = slCount(b->sharers);
    }
}

void fillInScores(struct clonePair *pairList)
/* Fill in scores. */
{
struct clonePair *pair;
struct oogClone *a, *b;
int q;

for (pair = pairList; pair != NULL; pair = pair->next)
    {
    if (pair->aUses == 2 && pair->bUses == 2)
	q = 1;
    else
	{
	q = max(pair->aUses, pair->bUses);
	if (q > 2)
	    q = q*q;
	}
    pair->score = (double)pair->overlap/(double)q;
    }
}

int cmpPairScore(const void *va, const void *vb)
/* Compare two clonePair. */
{
const struct clonePair *a = *((struct clonePair **)va);
const struct clonePair *b = *((struct clonePair **)vb);
float dif = b->score - a->score;

if (dif > 0)
    return 1;
else if (dif < 0)
    return -1;
else
    return 0;
}


void dumpBarge(struct barge *b)
/* Write out barge to log file. */
{
struct dlNode *node;
struct oogClone *clone;
logg("barge %d :", b->id);
for (node = b->cloneList->head; node->next != NULL; node = node->next)
    {
    clone = node->val;
    logg(" %s", clone->name);
    }
logg("\n");
}

static int nextBargeId;

struct barge *emptyBarge(struct dlList *bargeList)
/* Return a new empty barge. */
{
struct barge *barge;
AllocVar(barge);
barge->id = ++nextBargeId;
barge->cloneList = newDlList();
barge->listNode = dlAddValTail(bargeList, barge);
return barge;
}

boolean tooManyConflicts(struct barge *aBarge, struct barge *bBarge, int maxConflict)
/* Return TRUE if conflict count exceeds tolerances. */
{
#ifdef SOON
if (stsMaxVariance == 0)
    return FALSE;
if (aBarge->stsCertainty > maxConflict &&  bBarge->stsCertainty > maxConflict)
    return TRUE;
#endif /* SOON */
return FALSE;
}

boolean stsRangesClose(struct barge *aBarge, struct barge *bBarge, int maxConflict)
/* Return TRUE if STS looks compatible with range. */
{
#ifdef SOON
char *chromA = aBarge->stsChrom;
char *chromB = bBarge->stsChrom;
float startA = aBarge->stsStart, endA = aBarge->stsEnd;
float startB = bBarge->stsStart, endB = bBarge->stsEnd;
float s, e;

if (chromA == NULL || chromB == NULL)
    return TRUE;
if (!sameString(chromA, chromB))
    {
    if (tooManyConflicts(aBarge, bBarge, maxConflict))
	{
	logg("sts range chromosome mismatch %s %s\n", chromA, chromB);
	return FALSE;
	}
    }
startB -= stsMaxVariance;
endB += stsMaxVariance;
s = max(startA, startB);
e = min(endA, endB);
if (s > e)
    {
    if (tooManyConflicts(aBarge, bBarge, maxConflict))
	{
	logg("sts ranges aren't close %f-%f and %f-%f\n", startA, endA, startB, endB);
	return FALSE;
	}
    }
#endif /* SOON */
return TRUE;
}

struct barge *bargeOfOne(struct oogClone *clone, struct dlList *bargeList)
/* Return singleton barge. */
{
struct barge *barge = emptyBarge(bargeList);
clone->bargeNode = dlAddValTail(barge->cloneList, clone);
clone->barge = barge;
#ifdef SOON
if (clone->stsChrom)
    {
    barge->stsStart = barge->stsEnd = clone->stsPos;
    barge->stsChrom = clone->stsChrom;
    barge->stsCertainty = 1;
    }
#endif /* SOON */
return barge;
}

void makeSingletonBarges(struct oogClone *cloneList, struct dlList *bargeList)
/* Add in barges that don't connect to anything. */
{
struct oogClone *clone;
for (clone = cloneList; clone != NULL; clone = clone->next)
     if (clone->barge == NULL)
         bargeOfOne(clone, bargeList);
}

struct barge *bargeFromPair(struct dlList *bargeList, struct oogClone *a, struct oogClone *b)
/* Create a barge with just the pair a/b. */
{
struct barge *barge;
struct oogClone *stsClone = NULL;

logg("bargeFromPair(%s %s)\n", a->name, b->name);
#ifdef SOON
if (!stsClose(a->stsChrom, a->stsPos, a->stsPos, b->stsChrom, 
	b->stsPos, stsMaxConflict))
    return NULL;
#endif
barge = emptyBarge(bargeList);
#ifdef SOON 
if (a->stsChrom && b->stsChrom)
    {
    if (stsClose(a->stsChrom, a->stsPos, a->stsPos, b->stsChrom, 
	    b->stsPos, 0))
	{
	barge->stsChrom = a->stsChrom;
	if (a->stsPos < b->stsPos)
	    {
	    barge->stsStart = a->stsPos;
	    barge->stsEnd = b->stsPos;
	    }
	else
	    {
	    barge->stsStart = b->stsPos;
	    barge->stsEnd = a->stsPos;
	    }
	barge->stsCertainty = 2;
	}
    }
else 
    {
    if (a->stsChrom)
	stsClone = a;
    else if (b->stsChrom)
	stsClone = b;
    if (stsClone)
	{
	barge->stsChrom = stsClone->stsChrom;
	barge->stsStart = barge->stsEnd = stsClone->stsPos;
	barge->stsCertainty = 1;
	}
    }
#endif /* SOON */
a->bargeNode = dlAddValTail(barge->cloneList, a);
b->bargeNode = dlAddValTail(barge->cloneList, b);
a->barge = b->barge = barge;
return barge;
}

void fixCloneBargePointers(struct barge *barge)
/* Fix up all of the clones in barge->cloneList so that they point back to
 * this barge. */
{
struct dlNode *node;
struct oogClone *clone;
for (node = barge->cloneList->head; node->next != NULL; node = node->next)
    {
    clone = node->val;
    clone->barge = barge;
    clone->bargeNode = node;
    }
}

struct barge *pdupeBarge(struct barge *barge)
/* Return a partial duplicate of barge. This is done to explore if two
 * barges can be merged. This isn't completely independent of the
 * source barge, so be careful...  In particular the barge->cloneList->barge
 * pointers of the old barge get changed.  You will need to dispose of either
 * the original or the copy, and do a fixCloneBargePointers to restore things
 * to a normal state. */
{
struct barge *dupe;
struct dlList *oldList, *newList;
struct dlNode *node;

AllocVar(dupe);
*dupe = *barge;
dupe->cloneList = newList = newDlList();
oldList = barge->cloneList;
for (node = oldList->head; node->next != NULL; node = node->next)
    {
    dlAddValTail(newList, node->val);
    }
dupe->listNode = NULL;
dupe->usedPairs = NULL;
fixCloneBargePointers(dupe);
return dupe;
}

struct barge unpdupeBarge(struct barge **pBarge)
/* Free up a partially duped barge. */
{
struct barge *barge = *pBarge;
if (barge != NULL)
    freeDlList(&barge->cloneList);
freez(pBarge);
}


struct barge *extendBarge(struct barge *barge, struct oogClone *x, struct oogClone *b, 
	struct hash *looseHash)
/* Try to insert x into barge that already includes b. X must end up to one side or other of b.*/
{
struct oogClone *a = NULL, *c = NULL;
int abOverlap = 0, bcOverlap = 0;
int axOverlap = 0, bxOverlap = 0, cxOverlap = 0;
struct dlNode *node;
int nextTendency = 0, prevTendency = 0;
int xSize = x->size;
int maxErr = xSize/10;

logg("extendBarge(%s %s)\n", x->name, b->name);
dumpBarge(barge);
#ifdef SOON
if (!stsClose(barge->stsChrom, barge->stsStart, barge->stsEnd, x->stsChrom, 
	x->stsPos, stsMaxConflict))
    return NULL;
#endif /* SOON */
bxOverlap = lookupOverlap(looseHash, b, x);
node = b->bargeNode->prev;
if (node->prev != NULL)
    a = node->val;
node = b->bargeNode->next;
if (node->next != NULL)
    c = node->val;
if (a != NULL)
    {
    axOverlap = lookupOverlap(looseHash, a, x);
    abOverlap = lookupOverlap(looseHash, a, b);
    prevTendency = axOverlap - abOverlap;
    }
if (c != NULL)
    {
    cxOverlap = lookupOverlap(looseHash, c, x);
    bcOverlap = lookupOverlap(looseHash, b, c);
    nextTendency = cxOverlap - bcOverlap;
    }
if (prevTendency < -maxErr && nextTendency < -maxErr)
    {
    logg("Inconsistent tendencies\n");
    return NULL;
    }
if (prevTendency > nextTendency)
    {
    if (a  != NULL && axOverlap <= 0)
	{
	logg("Inconsistent axOverlap\n");
	return NULL;
	}
    x->bargeNode = dlAddValBefore(b->bargeNode, x);
    }
else
    {
    if (c  != NULL && cxOverlap <= 0)
	{
	logg("Inconsistent cxOverlap\n");
	return NULL;
	}
    x->bargeNode = dlAddValAfter(b->bargeNode, x);
    }
x->barge = barge;
#ifdef SOON
if (x->stsChrom)
    {
    if (barge->stsChrom == NULL)
	{
	barge->stsChrom = x->stsChrom;
	barge->stsStart = barge->stsEnd = x->stsPos;
	barge->stsCertainty = 1;
	}
    else
	{
	if (stsClose(barge->stsChrom, barge->stsStart, barge->stsEnd, x->stsChrom, 
		x->stsPos, 0))
	    {
	    if (x->stsPos < barge->stsStart)
		barge->stsStart = x->stsPos;
	    if (x->stsPos > barge->stsEnd)
		barge->stsEnd = x->stsPos;
	    barge->stsCertainty += 1;
	    }
	else
	    {
	    barge->stsCertainty -= 1;
	    }
	}
    }
#endif /* SOON */
logg("   ");
dumpBarge(barge);
return barge;
}

struct barge *wiggleIntoBarge(struct barge *barge, struct oogClone *x, struct oogClone *b, 
	struct hash *looseHash, int wiggleRoom)
/* Insert x into barge containing b somewhere in the near neighborhood of b. */
{
int maxOverlap = lookupOverlap(looseHash, x, b);
int prevOverlap = maxOverlap, nextOverlap = maxOverlap;
struct oogClone *maxOverlapClone = b;
int overlap, i;
struct oogClone *clone;
struct dlNode *prevNode, *nextNode;

prevNode = b->bargeNode->prev;
nextNode = b->bargeNode->next;
for (i=0; i<wiggleRoom; ++i)
    {
    if (nextNode != NULL && nextNode->next != NULL)
	{
	clone = nextNode->val;
	overlap = lookupOverlap(looseHash, x, clone);
	if (overlap < nextOverlap)
	    {
	    nextNode = NULL;
	    }
	else
	    {
	    if (overlap > maxOverlap)
		{
		maxOverlap = overlap;
		maxOverlapClone = clone;
		}
	    nextNode = nextNode->next;
	    }
	}
    if (prevNode != NULL && prevNode->prev != NULL)
	{
	clone = prevNode->val;
	overlap = lookupOverlap(looseHash, x, clone);
	if (overlap < prevOverlap)
	    {
	    prevNode = NULL;
	    }
	else
	    {
	    if (overlap > maxOverlap)
		{
		maxOverlap = overlap;
		maxOverlapClone = clone;
		}
	    prevNode = prevNode->prev;
	    }
	}
    }
return extendBarge(barge, x, maxOverlapClone, looseHash);
}

int cmpClonePos(const void *va, const void *vb)
/* Compare two clones. */
{
const struct oogClone *a = *((struct oogClone **)va);
const struct oogClone *b = *((struct oogClone **)vb);
return (a->bargePos - b->bargePos);
}

void sortAllClonesByPos(struct dlList *bargeList)
/* Sort all clones by their position in barge. */
{
struct dlNode *bNode;
struct barge *barge;

for (bNode = bargeList->head; bNode->next != NULL; bNode = bNode->next)
    {
    barge = bNode->val;
    dlSort(barge->cloneList, cmpClonePos);
    }
}


int bargeSize(struct barge *barge)
/* Figure out size of barge. */
{
struct dlNode *node;
int end, max=0;

for (node = barge->cloneList->head; node->next != NULL; node = node->next)
    {
    struct oogClone *clone = node->val;
    end = clone->bargePos + clone->size;
    if (end > max) max = end;
    }
return max;
}



boolean doMostOfMerge(struct barge *barge, struct oogClone *a, struct oogClone *b, struct hash *looseHash)
/* Merge clone a and other clones in a's barge into barge, starting at b. 
 * There is some truly scary sh*t happening with lists and pointers here so
 * that we can back out of the merge if it doesn't work out. */
{
struct dlNode *node, *prevNode = a->bargeNode->prev;
struct oogClone *clone, *overlapClone = b;
for (node = a->bargeNode; node->next != NULL; node = node->next)
    {
    clone = node->val;
    if (!wiggleIntoBarge(barge, clone, overlapClone, looseHash, 5))
	return FALSE;
    overlapClone = clone;
    }
overlapClone = a;
for (node = prevNode; node->prev != NULL; node = node->prev)
    {
    clone = node->val;
    if (!wiggleIntoBarge(barge, clone, overlapClone, looseHash, 5))
	return FALSE;
    overlapClone = clone;
    }
return TRUE;
}


boolean trialMerge(struct dlList *bargeList, struct oogClone *a, struct oogClone *b, struct hash *looseHash)
/* Merge a's barge into b's barge.  a and b overlap. Returns FALSE if can't 
 * There is some truly scary sh*t happening with list and pointers here so
 * that we can back out of the merge if it doesn't work out. */
{
struct barge *aBarge = a->barge;
struct barge *bBarge = b->barge;
struct barge *newBarge = pdupeBarge(bBarge);
struct dlNode *node;
struct oogClone *clone;

if (doMostOfMerge(newBarge, a, b, looseHash))
    {
    logg("after merge ");
    dumpBarge(newBarge);
    newBarge->usedPairs = slCat(aBarge->usedPairs, bBarge->usedPairs);
    newBarge->listNode = dlAddValAfter(bBarge->listNode, newBarge);
    fixCloneBargePointers(newBarge);
#ifdef SOON
    if (aBarge->stsChrom && bBarge->stsChrom)
	{
	if (stsRangesClose(aBarge, bBarge, 0))
	    {
	    newBarge->stsStart = min(aBarge->stsStart, bBarge->stsStart);
	    newBarge->stsEnd = max(aBarge->stsEnd, bBarge->stsEnd);
	    newBarge->stsCertainty = aBarge->stsCertainty + bBarge->stsCertainty;
	    }
	else
	    {
	    if (aBarge->stsCertainty > bBarge->stsCertainty)
		{
		newBarge->stsStart = aBarge->stsStart;
		newBarge->stsEnd = aBarge->stsEnd;
		newBarge->stsChrom = aBarge->stsChrom;
		newBarge->stsCertainty = aBarge->stsCertainty - bBarge->stsCertainty;
		}
	    else
		{
		newBarge->stsCertainty = bBarge->stsCertainty - aBarge->stsCertainty;
		}
	    }
	}
    else if (aBarge->stsChrom)
	{
	newBarge->stsStart = aBarge->stsStart;
	newBarge->stsEnd = aBarge->stsEnd;
	newBarge->stsChrom = aBarge->stsChrom;
	newBarge->stsCertainty = aBarge->stsCertainty;
	}
#endif /* SOON */
    dlRemove(bBarge->listNode);
    freez(&bBarge->listNode);
    unpdupeBarge(&bBarge);

    dlRemove(aBarge->listNode);
    freez(&aBarge->listNode);
    unpdupeBarge(&aBarge);
    return TRUE;
    }
else
    {
    fixCloneBargePointers(aBarge);
    fixCloneBargePointers(bBarge);
    unpdupeBarge(&newBarge);
    return FALSE;
    }
}

struct barge *mergeBarges(struct dlList *bargeList, struct oogClone *a, struct oogClone *b, 
	struct hash *looseHash)
/* Merge two barges if possible.  Return merged barge if possible, NULL otherwise. */
{
struct barge *aBarge = a->barge;
struct barge *bBarge = b->barge;

logg("mergeBarges(%s %s)\n", a->name, b->name);
    {
    logg(" ");
    dumpBarge(aBarge);
    logg(" ");
    dumpBarge(bBarge);
    }

#ifdef SOON
if (!stsRangesClose(aBarge, bBarge, stsMaxConflict))
    return NULL;
#endif

if (!trialMerge(bargeList, a, b, looseHash))
    {
    logg("Trial merge between %s (%d) and %s (%d) didn't work\n", 
    	a->name, a->barge->id, b->name, b->barge->id);
    return NULL;
    }
else
    return b->barge;
}

struct barge *fitIntoBarges(struct dlList *bargeList, struct clonePair *pair, struct hash *looseHash)
/* Fit pair into barge list if possible. */
{
struct oogClone *a = pair->a, *b = pair->b;
struct barge *barge;

logg("fitIntoBarges(%f %s %s %d\n", pair->score, pair->a->name, pair->b->name, pair->overlap);
if (a->barge == NULL && b->barge == NULL)
    {
    barge = bargeFromPair(bargeList, a, b);
    }
else if (a->barge == NULL)
    {
    barge = extendBarge(b->barge, a, b, looseHash);
    }
else if (b->barge == NULL)
    {
    barge = extendBarge(a->barge, b, a, looseHash);
    }
else if (a->barge == b->barge)
    {
    logg("Redundant overlap between %s and %s in barge %d\n", a->name, b->name, a->barge->id);
    barge = a->barge;
    }
else
    {
    barge = mergeBarges(bargeList, a, b, looseHash);
    }
return barge;
}


void markEnclosed(struct oogClone *cloneList, int maxOverlap, struct hash *looseHash)
/* Set flag on clones that look to be enclosed by another. */
{
struct oogClone *a, *b;
int overlap;
int aPpt, bPpt;

for (a = cloneList; a != NULL; a = a->next)
    a->isEnclosed = FALSE;
for (a = cloneList; a != NULL; a = a->next)
    {
    aPpt = a->maxOverlapPpt;
    if (aPpt > maxOverlap)
	{
	b = a->maxOverlapClone;
	overlap = lookupOverlap(looseHash, a, b);
	bPpt = round(1000.0*overlap/b->size);
	if (aPpt > bPpt)
	    a->isEnclosed = TRUE;
	else if (aPpt == bPpt && strcmp(a->name, b->name) < 0)
	    a->isEnclosed = TRUE;
	}
    }
}

void buryClone(struct oogClone *small, struct oogClone *big, struct dlList *bargeList, struct hash *looseHash)
/* Bury small clone in middle of big clone. */
{
struct barge *barge = big->barge;

if (barge == NULL)
    barge = bargeOfOne(big, bargeList);
if (small->barge == barge)
    return;
if (small->barge == NULL)
    {
    small->barge = barge;
    small->bargeNode = dlAddValAfter(big->bargeNode, small);
    small->bargePos = big->bargePos + (big->size - small->size)/2;
    }
else
    {
    int pos = big->bargePos + (big->size - small->size)/2;
    struct barge *sb = small->barge;
    struct dlNode *bigNode = big->bargeNode;
    struct dlNode *smallNode;
    struct oogClone *smallClone;

    while ((smallNode = dlPopHead(sb->cloneList)) != NULL)
	{
	smallClone = smallNode->val;
	dlAddAfter(bigNode, smallNode);
	smallClone->barge = barge;
	smallClone->bargePos += pos;
	bigNode = smallNode;
	}
    dlRemove(sb->listNode);
    barge->usedPairs = slCat(sb->usedPairs, barge->usedPairs);
    }
}

void placeEnclosedClones(struct oogClone *cloneList, struct clonePair **pPairList, 
	struct dlList *bargeList, struct hash *looseHash, int minPpt)
/* Place clones that are surrounded by another. */
{
struct clonePair *pair, *next, *leftovers = NULL;
struct oogClone *a, *b;
int overlap;

for (pair = *pPairList; pair != NULL; pair = next)
    {
    next = pair->next;
    a = pair->a;
    b = pair->b;
    if (a->maxOverlapClone == b && a->isEnclosed)
	{
	buryClone(a, b, bargeList, looseHash);
	slAddHead(&a->barge->usedPairs, pair);
	}
    else if (b->maxOverlapClone == a && b->isEnclosed)
	{
	buryClone(b, a, bargeList, looseHash);
	slAddHead(&b->barge->usedPairs, pair);
	}
    else
	{
	slAddHead(&leftovers, pair);
	}
    }
sortAllClonesByPos(bargeList);
*pPairList = leftovers;
}

void positionClones(struct dlList *bargeList, struct hash *looseHash)
/* Position clones relative to start of barge. */
{
struct dlNode *bNode;
struct dlNode *cNode, *ncNode;
struct barge *barge;
struct oogClone *clone, *nextClone;
int offset;

for (bNode = bargeList->head; bNode->next != NULL; bNode = bNode->next)
    {
    barge = bNode->val;
    offset = 0;
    for (cNode = barge->cloneList->head; cNode->next != NULL; cNode = cNode->next)
	{
	clone = cNode->val;
	clone->bargePos = offset;
	offset += clone->size;
	ncNode = cNode->next;
	if (ncNode->next != NULL)
	    {
	    nextClone = ncNode->val;
	    offset -= lookupOverlap(looseHash, clone, nextClone);
	    }
	}
    }
}

void glomBarges(struct oogClone *cloneList, struct clonePair **pPairList, 
	struct dlList *bargeList, struct hash *looseHash,
	int minScore, int maxOverlap, float maxStsDistance, int maxStsConflict)
/* Glom together clones into barges. Go through list of clone pairings in order
 * and if pairing is consistent with barges built so far use it to extend
 * barge collection. */
{
struct clonePair *leftovers = NULL, *next, *pair;
struct clonePair *rejects = NULL;
struct oogClone *a, *b;
int overlap;
int aSize, bSize;
struct barge *barge;
int rejectCount = 0, usedCount = 0, deferredCount = 0;

markEnclosed(cloneList, maxOverlap, looseHash);
for (pair = *pPairList; pair != NULL; pair = next)
    {
    next = pair->next;
    logg("pair %1.2f %s %s %d\n", pair->score, pair->a->name, pair->b->name, pair->overlap);
    if (pair->score >= minScore)
	{
	a = pair->a;
	b = pair->b;
	if (!a->isEnclosed && !b->isEnclosed)
	    {
	    if ((barge = fitIntoBarges(bargeList, pair, looseHash)) != NULL)
		{
		slAddHead(&barge->usedPairs, pair);
		++usedCount;
		}
	    else
		{
		slAddHead(&rejects, pair);
		++rejectCount;
		}
	    pair = NULL;
	    }
	else
	    {
	    logg(" rejected on overlap (%d %d %d)\n", 
	    	a->maxOverlapPpt, b->maxOverlapPpt, maxOverlap);
	    }
	}
    else
	{
	logg(" rejected on score (%d %d)\n", round(pair->score), minScore);
	}
    if (pair != NULL)
	{
	++deferredCount;
	slAddHead(&leftovers, pair);
	}
    }
slReverse(&leftovers);
// slReverse(&rejects);
// leftovers = slCat(leftovers, rejects);
*pPairList = leftovers;
positionClones(bargeList, looseHash);
status("Used %d, rejected %d, deferred %d pairs\n", usedCount, rejectCount, deferredCount);
}


void makeEasyBarges(struct oogClone *cloneList, struct clonePair **pPairList, 
	struct dlList *bargeList, struct hash *looseHash)
/* Make the relatively easy barges. */
{
glomBarges(cloneList, pPairList, bargeList, looseHash, 3000, 850, 5, 0);
}

void makeMediumBarges(struct oogClone *cloneList, struct clonePair **pPairList, 
	struct dlList *bargeList, struct hash *looseHash)
/* Make the relatively easy barges. */
{
glomBarges(cloneList, pPairList, bargeList, looseHash, 1500, 900, 10, 0);
}

int hardMaxOverlap = 950;  /* Maximum overlap in parts per thousand allowed between
                            * clones in last pass. */

void makeHardBarges(struct oogClone *cloneList, struct clonePair **pPairList, 
	struct dlList *bargeList, struct hash *looseHash)
/* Make the relatively easy barges. */
{
glomBarges(cloneList, pPairList, bargeList, looseHash, 0, hardMaxOverlap, 20, 0);
}

void flipBarge(struct barge *barge)
/* Flip order of cloneList. */
{
struct dlList *oldList = barge->cloneList;
struct dlList *newList = newDlList();
struct dlNode *node;
int bSize = bargeSize(barge);
int cSize;
struct oogClone *clone;
int cEnd;

while ((node = dlPopHead(oldList)) != NULL)
    {
    clone = node->val;
    cEnd = clone->bargePos + clone->size;
    clone->bargePos = bSize - cEnd;
    dlAddHead(newList, node);
    }
freeDlList(&oldList);
barge->cloneList = newList;
}

boolean needsFlipping(struct barge *barge)
/* Return TRUE if STS data indicates barge needs flipping. */
{
int tendency = 0;
boolean first = TRUE;
float pos, lastPos;
struct dlNode *cNode;
struct oogClone *clone;

for (cNode = barge->cloneList->head; cNode->next != NULL; cNode = cNode->next)
    {
    clone = cNode->val;
    if (first)
	{
	first = FALSE;
	lastPos = clone->mapPos;
	}
    else
	{
	pos = clone->mapPos;
	tendency += pos - lastPos;
	lastPos = pos;
	}
    }
return tendency < 0;
}

void flipBargesAsNeeded(struct dlList *bargeList)
/* Look if map positions make barge want to flip. */
{
struct dlNode *bNode;
struct barge *barge;

for (bNode = bargeList->head; bNode->next != NULL; bNode = bNode->next)
    {
    barge = bNode->val;
    if (needsFlipping(barge))
	flipBarge(barge);
    }
sortAllClonesByPos(bargeList);
}


static struct aliFilter tightFilter =  {200,10,100,500,1500};
static struct aliFilter mediumFilter = {400,20,40,200,1000};
static struct aliFilter looseFilter =  {600,20,25,100,100}; 

void writeBarges(char *fileName, struct dlList *bargeList, struct hash *looseHash)
/* Write out barges. */
{
FILE *f = mustOpen(fileName, "w");
struct dlNode *bargeNode, *cloneNode, *nextNode;
struct barge *barge;
struct oogClone *clone, *nextClone, *maxClone;
char *maxName;
int overlap;

for (bargeNode = bargeList->head; bargeNode->next != NULL; bargeNode = bargeNode->next)
    {
    barge = bargeNode->val;
    fprintf(f, "barge %d clones %d \n",
	barge->id, dlCount(barge->cloneList));
    for (cloneNode = barge->cloneList->head; cloneNode->next != NULL; cloneNode = cloneNode->next)
	{
	overlap = 0;
	clone = cloneNode->val;
	nextNode = cloneNode->next;
	if (nextNode->next != NULL)
	    {
	    nextClone = nextNode->val;
	    overlap = lookupOverlap(looseHash, clone, nextClone);
	    }
	maxName = "n/a";
	maxClone = clone->maxOverlapClone;
	if (maxClone != NULL) maxName = maxClone->name;
	fprintf(f, "%d\t%s %6d %6d  (%-8s %6d %4d)", clone->bargePos, clone->name, clone->size, overlap, 
		maxName, round(0.001*clone->maxOverlapPpt*clone->size), clone->maxOverlapPpt);
	fprintf(f, "\n");
	}
    fprintf(f, "\n");
    }
fclose(f);
}

struct barge *makeBargeList(struct oogClone *cloneList,  
    struct psl *pslList, struct hash *fragHash, struct hash *cloneHash, struct hash *looseHash)
/* Create a list of barges out of connected fragments, but don't yet order
 * clones inside barges. */
{
struct clonePair *ocp, *looseList, *tightList;
struct hash *tightHash = newHash(0);
struct dlList *bargeList = newDlList();
struct barge *slBargeList;

tightList = findOverlappingClones(cloneList, &tightFilter, pslList, 
	fragHash, cloneHash, tightHash);
saveOcp(tightFileName, tightList);
looseList = findOverlappingClones(cloneList, &looseFilter, pslList, 
	fragHash, cloneHash, looseHash);
saveOcp(looseFileName, looseList);

sharesFromPairs(tightList);
countUses(tightList);
fillInScores(tightList);
findMaxOverlaps(cloneList, looseHash);
slSort(&tightList, cmpPairScore);

makeEasyBarges(cloneList, &tightList, bargeList, looseHash);
status("Made %d easy barges %d pairs left\n", dlCount(bargeList), slCount(tightList));
makeMediumBarges(cloneList, &tightList, bargeList, looseHash);
status("Made %d medium barges %d pairs left\n", dlCount(bargeList), slCount(tightList));
makeHardBarges(cloneList, &tightList, bargeList, looseHash);
status("Made %d hard barges %d pairs left\n", dlCount(bargeList), slCount(tightList));
placeEnclosedClones(cloneList, &tightList, bargeList, looseHash, hardMaxOverlap);
flipBargesAsNeeded(bargeList);
makeSingletonBarges(cloneList, bargeList);
    {
    writeBarges("final.foo", bargeList, looseHash);
    }
freeHash(&tightHash);
slBargeList = dlListToSlList(bargeList);
freeDlList(&bargeList);
return slBargeList;
}



int calcBargeMapPos(struct barge *barge)
/* Calculate barge map position by average map position of it's clones. */
{
int cloneCount = 0;
int totalPos = 0;
struct dlNode *node;
struct oogClone *clone;

for (node = barge->cloneList->head; node->next != NULL; node = node->next)
    {
    clone = node->val;
    totalPos += clone->mapPos;
    ++cloneCount;
    }
return totalPos/cloneCount;
}

int cmpBargeMapPos(const void *va, const void *vb)
/* Compare two oogFrags to sort by default position . */
{
const struct barge *a = *((struct barge **)va);
const struct barge *b = *((struct barge **)vb);
return a->mapPos - b->mapPos;
}

void linkRaftsToBarges(struct raft *raftList)
/* Link rafts to barges. */
{
struct raft *raft;
struct oogClone *clone;
for (raft = raftList; raft != NULL; raft = raft->next)
    {
    clone = raft->fragList->frag->clone;
    assert(clone->barge != NULL);
    raft->barge = clone->barge;
    }
}

struct barge *makeBarges(struct oogClone *cloneList, struct raft *raftList, struct psl *pslList,
	 struct hash *fragHash, struct hash *cloneHash, struct hash *ocpHash)
/* Make barges including ordering clones inside them. */
{
struct barge *bargeList, *barge;
struct raft *raft;
struct oogClone *clone;

bargeList = makeBargeList(cloneList, pslList, fragHash, cloneHash, ocpHash);
status("Got %d barges\n", slCount(bargeList));
for (barge = bargeList; barge != NULL; barge = barge->next)
    {
    barge->mapPos = calcBargeMapPos(barge);
    }
uglyf("Past loop A\n");
slSort(&bargeList, cmpBargeMapPos);
uglyf("Past sort A\n");
uglyf("Got to end of makeBarges\n");
return bargeList;
}

void saveBarges(char *fileName, struct barge *bargeList, struct diGraph *bargeGraph, struct hash *ocpHash)
/* Save barge list to file. */
{
FILE *f = mustOpen(fileName, "w");
struct barge *barge, *nextBarge;
struct oogClone *clone, *nextClone;
struct dlNode *node, *nextNode;

fprintf(f, "Barge (Connected Clone) File ooGreedy Version %d\n\n", version);
fprintf(f, "start\taccession\tsize\toverlap\n");
fprintf(f, "--------------------------------------\n");
for (barge = bargeList; barge != NULL; barge = barge->next)
    {
    for (node = barge->cloneList->head; node->next != NULL; node = node->next)
	{
	int overlap = 0;
	clone = node->val;
	nextNode = node->next;
	if ((nextNode->next) != NULL)
	    {
	    nextClone = nextNode->val;
	    overlap = findHashedOverlap(clone->name, nextClone->name, ocpHash);
	    }
	fprintf(f, "%6d\t%s\t%6d\t%6d\n",
	    clone->bargePos, clone->name, clone->size, overlap);
	}
    if ((nextBarge = barge->next) != NULL)
	{
	if (dgDirectlyFollows(bargeGraph, barge->graphNode, nextBarge->graphNode))
	    {
	    fprintf(f, "----- bridged gap -----\n");
	    }
	else
	    {
	    fprintf(f, "----- open gap -----\n");
	    }
	}
    }
fclose(f);
}


void calcPositions(struct barge *bargeList, struct oogFrag *fragList, struct diGraph *bargeGraph)
/* Calculate positions of barges, clones, frags. */
{
int bargeOff = 0;
struct barge *barge, *nextBarge;
struct oogFrag *frag;
struct oogClone *clone;
struct dlNode *node;

/* Set barge positions. */
for (barge = bargeList; barge != NULL; barge = barge->next)
    {
    barge->offset = bargeOff;
    bargeOff += barge->size;
    if ((nextBarge = barge->next) != NULL)
	{
	if (bargeGraph != NULL)
	    {
	    if (dgDirectlyFollows(bargeGraph, barge->graphNode, nextBarge->graphNode))
		bargeOff += bridgedBargeGapSize;
	    else
		bargeOff += unbridgedBargeGapSize;
	    }
	else
	    {
	    bargeOff += bridgedBargeGapSize;
	    }
	}
    }

/* Set clone positions. */
for (barge = bargeList; barge != NULL; barge = barge->next)
    {
    for (node = barge->cloneList->head; node->next != NULL; node = node->next)
	{
	clone = node->val;
	clone->contigPos = barge->offset + clone->bargePos;
	if (bargeGraph != NULL)
	    {
	    clone->startNode->priority = clone->contigPos;
	    clone->endNode->priority = clone->contigPos + clone->size;
	    }
	}
    }

/* Set frag positions. */
for (frag = fragList; frag != NULL; frag = frag->next)
    {
    frag->defaultPos = frag->clone->contigPos + frag->clonePos;
    }
}

void dumpCcList(struct connectedComponent *ccList, FILE *f)
/* Write out info on connected component to file. */
{
struct connectedComponent *cc;
struct dgNodeRef *ref;
fprintf(f, "dumpCcList:\n");
for (cc = ccList; cc != NULL; cc = cc->next)
    {
    fprintf(f, "%d    ", cc->orientation); 
    for (ref = cc->subGraph; ref != NULL; ref = ref->next)
	{
	fprintf(f, "%s ", ref->node->name);
	}
    fprintf(f, "\n");
    }
}


boolean setCcOrientation(struct connectedComponent *cc)
/* Set cc->orientation to be consistent between subGraph and subLine
 * if possible. If not return FALSE. */
{
struct dgNodeRef *nRef;
struct raft *raft;
int initialOrientation;
struct mrnaConnectionRef *lRef;
struct mrnaConnection *mc;

/* Take care of singleton unoriented rafts. */
nRef = cc->subGraph;
raft = nRef->node->val;
if (raft->orientation == 0)
    {
    assert(nRef->next == NULL);
    assert(cc->subLine->next == NULL);
    raft->orientation = cc->subLine->mc->orientation;
    cc->orientation = 1;
    return TRUE;
    }

lRef = cc->subLine;
mc = lRef->mc;
initialOrientation = mc->orientation*mc->raft->orientation;
for (lRef = cc->subLine->next; lRef != NULL; lRef = lRef->next)
    {
    mc = lRef->mc;
    if (mc->orientation*mc->raft->orientation != initialOrientation)
	return FALSE;
    }
cc->orientation = initialOrientation;
return TRUE;
}


boolean flipConnectedRafts(struct diGraph *raftGraph, struct dgNodeRef *raftList)
/* Flip around all rafts in list.  Return FALSE if can't because would induce cycles. 
 */
{
struct dgEdgeRef *erList;
struct dgNodeRef *ref;
struct raft *raft;
boolean ok;

fprintf(logFile, "Flipping connected component of %d rafts\n", slCount(raftList));
	{
	for (ref = raftList; ref != NULL; ref = ref->next)
	    {
	    raft = ref->node->val;
	    fprintf(logFile, " %d", raft->id);
	    }
	fprintf(logFile, "\n", raft->id);
	}
erList = dgFindSubEdges(raftGraph, raftList);
dgSwapEdges(raftGraph, erList);
ok = !dgHasCycles(raftGraph);
if (ok)
    {
    for (ref = raftList; ref != NULL; ref = ref->next)
	{
	raft = ref->node->val;
	raft->orientation *= -1;
	}
    }
else
    {
    dgSwapEdges(raftGraph, erList);
    }
slFreeList(&erList);
return ok;
}

boolean forceForwardOrientation(struct diGraph *raftGraph, struct connectedComponent *ccList)
/* Flip around connected subgraph orientations if necessary. */
{
struct connectedComponent *cc;
struct mrnaConnectionRef *lRef;
for (cc = ccList; cc != NULL; cc = cc->next)
    {
    if (cc->orientation < 0)
	{
	if (!flipConnectedRafts(raftGraph, cc->subGraph))
	    return FALSE;
	cc->orientation = 0;
	for (lRef = cc->subLine; lRef != NULL; lRef = lRef->next)
	    {
	    lRef->mc->orientation *= -1;
	    }
	}
    }
return TRUE;
}

boolean wouldAddCycle(struct diGraph *graph, struct mrnaConnection *mcList)
/* Add bits of mcList not already in graph to graph.  See if it has
 * cycles, remove them. Report if has cycles. */
{
int mcCount = slCount(mcList);
struct mrnaConnection *mc;
static int mcAlloc = 0;
static bool *added;
int i;
boolean isCyclic;

if (mcAlloc < mcCount)
    {
    /* Make sure have room to track all. */
    mcAlloc = mcCount*2;
    ExpandArray(added, 0, mcAlloc);
    }
memset(added, 0, mcCount);
for (mc=mcList, i=0; mc->next != NULL; mc = mc->next, ++i)
    {
    if (!dgDirectlyFollows(graph, mc->raft->node, mc->next->raft->node))
	{
	added[i] = TRUE;
	dgConnect(graph, mc->raft->node, mc->next->raft->node);
	}
    }
isCyclic = dgHasCycles(graph);
for (mc=mcList, i=0; mc->next != NULL; mc = mc->next, ++i)
    {
    if (added[i])
	{
	dgDisconnect(graph, mc->raft->node, mc->next->raft->node);
	}
    }
return isCyclic;
}

boolean canAddMrnaLine(struct diGraph *graph, struct mrnaLine *ml, 
	struct connectedComponent *ccList)
/* Return TRUE if can add line to graph.  Reverse direction of mcList if need be
 * according to orientation.  Return orientation. */
{
struct mrnaConnection *mc;
short lineOrientation = 0;
short oneOrientation;
struct dgNode *a, *b;
struct connectedComponent *cc;

/* Don't bother with 0 and 1 element lists. */
if (ml->mcList == NULL || ml->mcList->next == NULL)
    return FALSE;

/* First check for conflicts in orientation. */
for (cc = ccList; cc != NULL; cc = cc->next)
    {
    if (!setCcOrientation(cc))
	{
	fprintf(logFile, "Line orientation conflict with graph and %s\n", ml->name);
	return FALSE;
	}
    }
if (!forceForwardOrientation(graph, ccList))
    return FALSE;
/* Check for cycles. */
if (wouldAddCycle(graph, ml->mcList))
    {
    fprintf(logFile, "Cycle conflict with graph and %s (%d)\n", ml->name,
	ml->score);
    return FALSE;
    }
return TRUE;
}

void addMrnaLine(struct diGraph *graph, struct mrnaLine *ml, 
	struct connectedComponent *ccList)
/* Add line (already checked not to have cycle or orientation conflicts) to graph. */
{
struct mrnaConnection *mc, *mcNext;
struct dgNode *a, *b;

for (mc = ml->mcList; (mcNext = mc->next) != NULL; mc = mcNext)
    {
    a = mc->raft->node;
    b = mcNext->raft->node;
    dgConnect(graph, a, b);
    }
}


void clearRaftCc(struct raft *raftList)
/* Clear the connected component pointers from raftList. */
{
struct raft *raft;
for (raft = raftList; raft != NULL; raft = raft->next)
    raft->cc = NULL;
}

struct connectedComponent *findConnectedComponentsThatTouchLine(struct diGraph *graph,
	struct raft *raftList, struct mrnaLine *ml)
/* Make up a connectedComponent list. */
{
struct connectedComponent *ccList = NULL, *cc;
struct mrnaConnection *mc;
struct mrnaConnectionRef *mcRef;
struct dgNodeRef *subGraph;

clearRaftCc(raftList);
dgClearConnFlags(graph);
for (mc = ml->mcList; mc != NULL; mc = mc->next)
    {
    if ((cc = mc->raft->cc) == NULL)
	{
	struct dgNodeRef *nr;
	AllocVar(cc);
	subGraph = dgFindNewConnectedWithVals(graph, mc->raft->node);
	cc->subGraph = subGraph;
	for (nr = cc->subGraph; nr != NULL; nr = nr->next)
	    {
	    struct raft *raft = nr->node->val;
	    raft->cc = cc;
	    }
	slAddHead(&ccList, cc);
	}
    AllocVar(mcRef);
    mcRef->mc = mc;
    slAddHead(&cc->subLine, mcRef);
    }
return ccList;
}




void ooDumpGraph(struct diGraph *dg, FILE *out)
/* Dump info on graph to output. */
{
struct dgNode *node;
struct dgConnection *con;
struct raft *raft;

fprintf(out, "raft\tori\tways out\n");
fprintf(out, "-------------------------\n");
for (node = dg->nodeList; node != NULL; node = node->next)
    {
    if (node->nextList == NULL && node->prevList == NULL)
	continue;
    raft = node->val;
    if (raft != NULL)
	{
	fprintf(out, "%s\t%c\t", node->name, orientChar(raft->orientation));
	}
    else
	{
	fprintf(out, "%s\t", node->name);
	}
    for (con = node->nextList; con != NULL; con = con->next)
	{
	fprintf(out, "%s,", con->node->name);
	}
    fprintf(out, "\n");
    }
}


void addToBargeGraph(struct diGraph *bargeGraph, struct mrnaLine *ml)
/* Add mrnaLine to bargeGraph.  Connect in both directions... */
{
struct barge *lastBarge = NULL, *barge;
struct mrnaConnection *mc;

for (mc = ml->mcList; mc != NULL; mc = mc->next)
    {
    barge = mc->raft->barge;
    if (lastBarge != NULL && barge != lastBarge)
	{
	dgConnect(bargeGraph, barge->graphNode, lastBarge->graphNode);
	dgConnect(bargeGraph, lastBarge->graphNode, barge->graphNode);
	}
    lastBarge = barge;
    }
}

void maybeLogGraph(struct diGraph *graph)
/* Dump graph to log if in verbose mode. */
{
#ifdef VERBOSE_LOG
ooDumpGraph(graph, logFile);
#endif
}

struct cloneEndPos
/* Info on the end of a clone. */
    {
    struct cloneEndPos *next;	/* Next in list. */
    struct oogClone *clone;	/* Clone this refers to. */
    int pos;                    /* Position (within barge) */
    boolean isEnd;              /* True if end rather than start of clone. */
    };

int cmpCloneEndPos(const void *va, const void *vb)
/* Compare two cloneEndPos to sort by pos. */
{
const struct cloneEndPos *a = *((struct cloneEndPos **)va);
const struct cloneEndPos *b = *((struct cloneEndPos **)vb);
return (a->pos - b->pos);
}

void addBargeToRaftGraph(struct diGraph *raftGraph, struct barge *barge,
	struct oogClone **pLastClone)
/* Add a barge to the raft graph. */
{
struct oogClone *clone;
struct oogClone *lastClone = *pLastClone;
struct oogClone *newLastClone;
struct cloneEndPos *cepList = NULL, *cep;
struct dgNode *node, *lastNode = NULL;
char nodeName[256];
struct dlNode *n;

/* Build up list of clone starts and ends sorted by position in barge. */
for (n = barge->cloneList->head; n->next != NULL; n = n->next)
    {
    AllocVar(cep);
    clone = n->val;
    cep->clone = clone;
    cep->pos = clone->bargePos;
    cep->isEnd = FALSE;
    slAddHead(&cepList, cep);
    AllocVar(cep);
    cep->clone = clone;
    cep->pos = clone->bargePos + clone->size;
    cep->isEnd = TRUE;
    slAddHead(&cepList, cep);
    }
slSort(&cepList, cmpCloneEndPos);

for (cep = cepList; cep != NULL; cep = cep->next)
    {
    clone = cep->clone;
    if (cep->isEnd)
	{
	sprintf(nodeName, "end-%s", clone->name);
	clone->endNode = node = dgAddNode(raftGraph, nodeName, NULL);
	}
    else
	{
	sprintf(nodeName, "start-%s", clone->name);
	clone->startNode = node = dgAddNode(raftGraph, nodeName, NULL);
	}
    if (lastNode != NULL)
	dgConnect(raftGraph, lastNode, node);
    lastNode = node;
    }
*pLastClone = clone;

if (lastClone != NULL)
    {
    clone = cepList->clone;
    dgConnect(raftGraph, lastClone->endNode, clone->startNode);
    }

slFreeList(&cepList);
}

struct oogClone *singleCloneForRaft(struct raft *raft)
/* If raft is made up of fragments from a single clone, return that
 * clone.  Otherwise return NULL. */
{
struct raftFrag *rf = raft->fragList;
struct oogClone *clone = rf->frag->clone;

for (rf = rf->next; rf != NULL; rf = rf->next)
    {
    if (rf->frag->clone != clone)
	return NULL;
    }
return clone;
}

struct barge *singleBargeForRaft(struct raft *raft)
/* Return barge for raft if all frags in raft belong to a single
 * barge, otherwise NULL. */
{
struct raftFrag *rf = raft->fragList;
struct barge *barge = rf->frag->clone->barge;

for (rf = rf->next; rf != NULL; rf = rf->next)
    {
    if (rf->frag->clone->barge != barge)
	return NULL;
    }
return barge;
}

boolean raftHitsClone(struct raft *raft, struct oogClone *clone)
/* Returns true if a fragment of raft is in clone. */
{
struct raftFrag *rf;

for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    if (rf->frag->clone == clone)
	return TRUE;
    }
return FALSE;
}

boolean cloneBoundsInBarge(struct raft *raft, struct oogClone **retStart, 
	struct oogClone **retEnd)
/* Return clones that bound raft if raft lies in a single barge.
 * (Otherwise return FALSE.) */
{
struct barge *barge;
struct oogClone *clone, *start = NULL, *end = NULL;
int maxEndPos = -0x3ffffff;
int endPos;
struct dlNode *node;

if ((barge = singleBargeForRaft(raft)) == NULL)
    return FALSE;
for (node = barge->cloneList->head; node->next != NULL; node = node->next)
    {
    clone = node->val;
    if (raftHitsClone(raft, clone))
	{
	if (start == NULL)
	    start = clone;
	endPos = clone->bargePos + clone->size;
	if (endPos > maxEndPos)
	    end = clone;
	}
    }
assert(start != NULL);
*retStart = start;
*retEnd = end;
return TRUE;
}

void addRaftToRaftGraph(struct diGraph *raftGraph, struct raft *raft)
/* Add a raft to raft graph - make sure it stays inside of involved
 * clones if at all possible. */
{
struct oogClone *clone, *startClone, *endClone;
struct dgNode *raftNode;

raft->node = raftNode = dgAddNumberedNode(raftGraph, raft->id, raft);
if ((clone = singleCloneForRaft(raft)) != NULL)
    {
    dgConnect(raftGraph, clone->startNode, raftNode);
    dgConnect(raftGraph, raftNode, clone->endNode);
    }
else if (cloneBoundsInBarge(raft, &startClone, &endClone))
    {
    dgConnect(raftGraph, startClone->startNode, raftNode);
    dgConnect(raftGraph, raftNode, endClone->endNode);
    }
else
    {
    warn("Raft %d lies in multiple barges", raft->id);
    }
}

void makeRaftGraphSkeleton(struct diGraph *raftGraph, struct barge *bargeList, 
	struct raft *raftList)
/* Make skeleton of raftGraph - dummy nodes for ends of each clone and 
 * real nodes for each raft. Constrain raft nodes to lie between appropriate
 * clone end nodes. */
{
struct barge *barge;
struct raft *raft;
struct oogClone *lastClone = NULL;

for (barge = bargeList; barge != NULL; barge = barge->next)
    addBargeToRaftGraph(raftGraph, barge, &lastClone);
for (raft = raftList; raft != NULL; raft = raft->next)
    addRaftToRaftGraph(raftGraph, raft);
slReverse(&raftGraph->nodeList);
}

void reverseMl(struct mrnaLine *ml)
/* Reverse order of mrnaLine. */
{
struct mrnaConnection *mc;

for (mc = ml->mcList; mc != NULL; mc = mc->next)
    mc->orientation = -mc->orientation;
slReverse(&ml->mcList);
}

void makeGraphs(struct raft *raftList, struct barge *bargeList, 
	struct mrnaLine *mrnaLineList, struct diGraph **retRaftGraph, 
	struct diGraph **retBargeGraph)
/* Make a graphs out of mRNA lines. */
{
struct mrnaLine *ml;
struct raft *raft;
struct barge *barge;
struct dgNode *node;
struct diGraph *raftGraph = dgNew();
struct diGraph *bargeGraph = dgNew();
struct connectedComponent *ccList, *cc;
boolean okOne;
int okCount = 0;
int conflictCount = 0;
boolean revMrna;


fprintf(logFile, "\nmakeGraphs\n");

/* Add skeleton of clones to raftGraph. */
makeRaftGraphSkeleton(raftGraph, bargeList, raftList);
ooDumpGraph(raftGraph, logFile);

/* Add all barges to barge graph. */
for (barge = bargeList; barge != NULL; barge = barge->next)
    barge->graphNode = node = dgAddNumberedNode(bargeGraph, barge->id, barge);
slReverse(&bargeGraph->nodeList);

/* Add in non-conflicting mrna lines one at a time. */
for (ml = mrnaLineList; ml != NULL; ml = ml->next)
    {
    okOne = FALSE;
    for (revMrna = FALSE; revMrna <= TRUE; ++revMrna)
	{
	ccList = findConnectedComponentsThatTouchLine(raftGraph, raftList, ml);
	dumpMrnaLine(ml, logFile);
	dumpCcList(ccList, logFile);
	if (canAddMrnaLine(raftGraph, ml, ccList))
	    {
	    addMrnaLine(raftGraph, ml, ccList);
	    addToBargeGraph(bargeGraph, ml);
	    dumpMrnaLine(ml, logFile);
	    fprintf(logFile, "Graph after addMrnaLine:\n");
	    maybeLogGraph(raftGraph);
	    okOne = TRUE;
	    break;
	    }
	reverseMl(ml);
	}
    if (okOne)
	++okCount;
    else
	{
	fprintf(logFile, "Couldn't add mrnaLine\n");
	++conflictCount;
	}
    fprintf(logFile, "\n");
    }
fprintf(logFile, "%d mrnaLines ok, %d conflicted\n", okCount, conflictCount);
fprintf(stdout, "%d mrnaLines ok, %d conflicted\n", okCount, conflictCount);
*retRaftGraph = raftGraph;
*retBargeGraph = bargeGraph;
}

void loadClone(char *faName, struct oogClone *clone, struct hash *fragHash, struct oogFrag **pFragList)
/* Position fragments within a clone based on position in
 * .fa file. */
{
FILE *f = mustOpen(faName, "r");
DNA *dna;
int sizeOne;
int totalSize = 0;
struct hashEl *hel;
struct oogFrag *frag;
struct dnaSeq *seq;
char *commentLine;

while (faReadNext(f, NULL, TRUE, &commentLine, &seq))
    {
    if ((hel = hashLookup(fragHash, seq->name)) != NULL)
	{
	warn("Duplicate %s, ignoring all but first\n", seq->name);
	}
    else
	{
	AllocVar(frag);
	slAddHead(pFragList, frag);
	hel = hashAdd(fragHash, seq->name, frag);
	frag->name = hel->name;
	frag->seq = seq;
	frag->clone = clone;
	frag->size = seq->size;
	frag->clonePos = totalSize;
	totalSize += seq->size;
	}
    }
clone->size = totalSize;
fclose(f);
}

int loadAllClones(char *infoName, char *genoList, 
	struct hash *fragHash, struct hash *cloneHash,
	struct oogFrag **pFragList, struct oogClone **pCloneList,
	char **retCtgName)
/* Consult info and geno list and associated fa files to build
 * default positions for fragments. */
{
struct lineFile *lf;
char *line;
int lineSize;
char *words[3];
int wordCount;
char **genoFiles;
int genoCount;
char *faFile;
char *acc;
int offset;
struct hashEl *hel;
struct oogClone *clone;
struct oogFrag *frag;
char *gBuf;
int i;
char dir[256], name[128],extension[64];
int endPos, lastEnd = 0;
int officialContigSize = 0;
struct slName *miaList = NULL, *mia;

/* Process geno.lst file. */
readAllWords(genoList, &genoFiles, &genoCount, &gBuf);
if (genoCount <= 0)
    errAbort("%s is empty\n", genoList);
for (i=0; i<genoCount; ++i)
    {
    faFile = genoFiles[i];
    splitPath(faFile, dir, name, extension);
    acc = name;
    if (fileExists(faFile))
	{
	if ((hel = hashLookup(cloneHash, acc)) != NULL)
	    {
	    warn("Duplicate %s in geno.lst, ignoring all but first", acc);
	    }
	else
	    {
	    AllocVar(clone);
	    slAddHead(pCloneList, clone);
	    hel = hashAdd(cloneHash, acc, clone);
	    clone->name = hel->name;
	    loadClone(faFile, clone, fragHash, pFragList);
	    }
	}
    else
	{
	warn("No sequence for %s", acc);
	mia = newSlName(acc);
	slAddHead(&miaList, mia);
	}
    }

/* Process info file. */
lf = lineFileOpen(infoName, TRUE);
if (!lineFileNext(lf, &line, &lineSize))
    errAbort("Empty info file %s\n", infoName);
if ((wordCount = chopLine(line, words)) < 2)
    errAbort("Bad header on %s\n", lf->fileName);
*retCtgName = cloneString(words[0]);
if (wordCount >= 3)
    officialContigSize = atoi(words[2]);
while (lineFileNext(lf, &line, &lineSize))
    {
    if ((wordCount = chopLine(line, words)) < 2)
	errAbort("Bad info file format line %d of %s\n", lf->lineIx, lf->fileName);
    if (!words[1][0] == '-' && !isdigit(words[1][0]))
	errAbort("Bad info file format line %d of %s\n", lf->lineIx, lf->fileName);
    acc = words[0];
    if ((hel = hashLookup(cloneHash, acc)) == NULL)
	{
	if (!slNameInList(miaList, acc))
	    {
	    fprintf(logFile, "Clone %s is in info but not geno.lst", acc);
	    }
	}
    else
	{
	clone = hel->val;
	clone->mapPos = atoi(words[1]) * 1000;
	if (wordCount >= 3)
	    clone->phase = atoi(words[2]);
	}
    }
lineFileClose(&lf);

for (frag = *pFragList; frag != NULL; frag = frag->next)
    {
    frag->defaultPos = frag->clone->mapPos + frag->clonePos;
    endPos = frag->defaultPos + frag->size;
    if (endPos > lastEnd)
	lastEnd = endPos;
    }
freeMem(gBuf);
slFreeList(&miaList);
slReverse(pFragList);
slReverse(pCloneList);
if (officialContigSize > 0)
    return officialContigSize;
else
    return lastEnd;
}

void oneRaftDefaultPosition(struct raft *raft)
/* Set position of raft to weighted average of position of fragments. */
{
double totalBases = 0;
double weightedTotal = 0;
struct raftFrag *rf;
struct oogFrag *frag;
double sizeOne;
double posOne;

for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    frag = rf->frag;
    sizeOne = frag->size;
    totalBases += sizeOne;
    posOne = frag->defaultPos + 0.5*sizeOne;
    weightedTotal += sizeOne*posOne;
    }
raft->node->priority = raft->defaultPos = round(weightedTotal/totalBases);
}

void setRaftDefaultPositions(struct raft *raftList)
/* Set position of all rafts to weighted average of position of fragments. */
{
struct raft *raft;
for (raft = raftList; raft != NULL; raft = raft->next)
    oneRaftDefaultPosition(raft);
}

void writeFragMap(char *fragMapName, struct oogFrag *fragList)
/* Write out default positions of frags. */
{
struct oogFrag *frag;
FILE *f = mustOpen(fragMapName, "w");
for (frag = fragList; frag != NULL; frag = frag->next)
    {
    fprintf(f, "%s\t%d\t%d\n", frag->name, frag->defaultPos, frag->size);
    }
fclose(f);
}

struct dgNodeRef *orderRafts(struct diGraph *graph)
/* Return ref list of ordered raft nodes. */
{
return dgConstrainedPriorityOrder(graph);
}


void writeOrderList(struct dgNodeRef *orderList, char *raftOrderName)
/* Write out raft id's in order to raftOrderName. */
{
struct dgNodeRef *ref;
struct raft *raft;
FILE *f = mustOpen(raftOrderName, "w");

for (ref = orderList; ref != NULL; ref = ref->next)
    {
    raft = ref->node->val;
    if (raft != NULL)
	{
	fprintf(f, "%d %c defaultPos %d\n", raft->id, 
	    (raft->orientation < 0 ? '-' : '+'),
	    raft->defaultPos);
	}
    }
fclose(f);
}

int raftFlipTendency(struct raft *raft)
/* Return how badly raft would like to fit to better
 * accomodate default coordinates. */
{
/* raft->fragList is sorted already, and needs to be for
 * this to work. */
int diffTotal = 0;
int diffOne;
struct raftFrag *rf, *rfNext;

for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    if ((rfNext = rf->next) != NULL)
	{
	diffOne = rfNext->frag->defaultPos - rf->frag->defaultPos;
	diffTotal += diffOne;
	}
    }
if (raft->orientation < 0)
    diffTotal = -diffTotal;
return diffTotal;
}

void flipNearDefaults(struct diGraph *graph)
/* Flip around connected components of graphs so as to make them
 * closer to default values. */
{
struct dgNodeRef *ccList, *cc, *ccNext;
struct raft *raft, *nextRaft;

dgClearConnFlags(graph);
while ((ccList = dgFindNextConnectedWithVals(graph)) != NULL)
    {
    int flipTendency = 0;
    int oneFlipTendency;
    int raftDistance;

    for (cc = ccList; cc != NULL; cc = cc->next)
	{
	raft = cc->node->val;
	if (raft != NULL)
	    {
	    if (raft->orientation == 0)
		raft->orientation = 1;
	    oneFlipTendency = raftFlipTendency(raft);
	    flipTendency += oneFlipTendency;
	    if ((ccNext = cc->next) != NULL)
		{
		nextRaft = ccNext->node->val;
		if (nextRaft != NULL)
		    {
		    raftDistance = nextRaft->defaultPos - raft->defaultPos;
		    if (dgPathExists(graph, ccNext->node, cc->node))
			{
			raftDistance = -raftDistance;
			}
		    flipTendency += 4*raftDistance;
		    }
		}
	    }
	}
    if (flipTendency < 0)
	{
	boolean ok;
	fprintf(logFile, "Flipping tendency = %d\n", flipTendency);
	ok = flipConnectedRafts(graph, ccList);
	fprintf(logFile, "Flipped %s\n", (ok ? "ok" : "not ok"));
	}
    slFreeList(&ccList);
    }
}

enum gapType
/* Type of gap this is. */
    {
    gtUnknown = 0,	/* Who knows what type? */
    gtFragment = 1,     /* Gap between fragments of a clone. */
    gtClone = 2,        /* Gap between clones. */
    gtContig = 3        /* Gap between contigs. */
    };
char *gapTypeNames[] = { "?", "fragment", "clone", "contig"};

enum bridgeType
/* Type of bridge covering gap. */
    {
    btUnknown = 0,     /* Who knows... */
    btYes = 1,         /* Some sort of bridge. */
    btNo = 2,          /* No bridge. */
    btMrna = 3,        /* Bridged by mRNA. */
    btEst = 4,         /* Bridged by EST. */
    btEstPair = 5,     /* Bridged by EST pair. */
    btBacEndBair = 6   /* Bridged by BAC end pair. */
    };
char *bridgeTypeNames[] = { "?", "yes", "no", "mRNA", "EST", "EST_pair", "BAC_end_pair"};

struct goldenPath
/* The way to patch together the sequence for this contig. */
    {
    struct goldenPath *next;    /* Next in list. */
    struct oogFrag *frag;	/* fragment or NULL if "just N's" */
    int fragStart;              /* Start in fragment.  If orientation -1 this is size-end.  */
    int fragEnd;                /* End in fragment.  If orientation -1 this is size-start. */
    short orientation;          /* Orientation of fragment. */
    enum gapType gapType;       /* Gap type (just for N's) */
    enum bridgeType bridgeType; /* Bridge type (just for N's) */
    };

struct raftPath
/* A golden path through one raft */
    {
    struct raftPath *next;     /* Next in list. */
    struct goldenPath *gpList; /* Golden path through this raft. */
    struct raft *raft;	       /* Which raft this is. */
    int start;	                /* Start in contig coordinates. */
    int size;                  /* size of raft. */
    };

void dumpGp(struct goldenPath *gp, FILE *f)
/* Write out golden path element. */
{
fprintf(f, "%d\t", gp->fragEnd - gp->fragStart);
if (gp->frag != NULL)
    {
    fprintf(f, "%s\t%c\t%d\t%d\n", gp->frag->name, (gp->orientation > 0 ? '+' : '-'),
	    gp->fragStart, gp->fragEnd);
    }
else
    {
    fprintf(f, "N\n");
    }
}

enum {bigEnoughSize = 20};

int firstBigBlock(unsigned *sizes, int count)
/* Find first big block in list, or if can't find one then first block. */
{
int i;
for (i=0; i<count; ++i)
    if (sizes[i] >= bigEnoughSize)
	return i;
return 0;
}

int lastBigBlock(unsigned *sizes, int count)
/* Find last big block in list or if can't find one then last block. */
{
int i = count;
while (--i >= 0)
    if (sizes[i] >= bigEnoughSize)
	return i;
return count-1;
}

int midBigBlock(unsigned *sizes, int count)
/* Find middle big block in list of if can't then just plain middle block. */
{
if (count == 1)
    return 0;
else
    {
    int half = count/2;
    int betterHalf = (count+1)/2;
    int i;

    for (i=0; i<=betterHalf; ++i)
	{
	int ix;
	ix = half + i;
	if (ix < count && sizes[ix] >= bigEnoughSize)
	    return ix;
	ix = half - i;
	if (ix >= 0 && sizes[ix] >= bigEnoughSize)
	    return ix;
	}
    return half;
    }
}

void forwardCopyBlocks(int blockCount, unsigned *sizes, unsigned *oSizes,
	unsigned *aStarts, unsigned *oaStarts, unsigned *bStarts, unsigned *obStarts)
/* Copy blocks without any swapping. */
{
int i;
for (i=0; i<blockCount; ++i)
    {
    oaStarts[i] = aStarts[i];
    obStarts[i] = bStarts[i];
    oSizes[i] = sizes[i];
    }
}

void reverseCopyBlocks(int blockCount, int aSize, int bSize,
	unsigned *sizes, unsigned *oSizes,
	unsigned *aStarts, unsigned *oaStarts, 
	unsigned *bStarts, unsigned *obStarts)
/* Copy blocks and swap. */
{
int i;
for (i=0; i<blockCount; ++i)
    {
    int revIx = blockCount-1-i;
    int blockSize = sizes[revIx];
    oaStarts[i] = aSize - (aStarts[revIx] + blockSize);
    obStarts[i] = bSize - (bStarts[revIx] + blockSize);
    oSizes[i] = blockSize;
    }
}

enum crossAt {crossUsual, crossMiddle, crossExtreme};

void findCrossover(struct oogEdge *edge,  struct oogFrag *aFrag, struct oogFrag *bFrag,
	struct goldenPath *a, struct goldenPath *b, enum crossAt atMiddle)
/* Figure out crossover point to get from frag to ext->other.  Use this and data in 'a'
 * to fill in 'b' and update ends of 'a' */
{
int aSize = aFrag->size, bSize = bFrag->size;
struct psl *psl = edge->psl;
int cbIx;
int cbSize;
int cbPt;
int aCross, bCross;
static int blockAlloc = 0;
int blockCount = psl->blockCount;
static unsigned *oaStarts = NULL, *obStarts = NULL, *oBlockSizes = NULL;
short aOrient = a->orientation, bOrient;
int aQual, bQual;
bool moreFromA;	/* Take more sequence from A? */

fprintf(logFile, "findCrossover %s(%d) %d %s(%d)\n", 
	aFrag->name, a->orientation, edge->orientation, bFrag->name, b->orientation);
fprintf(logFile, "psl: (%s %d %d of %d) %s (%s %d %d of %d)\n", 
       psl->qName, psl->qStart, psl->qEnd, psl->qSize, 
       psl->strand,
       psl->tName, psl->tStart, psl->tEnd, psl->tSize);
/* Make sure local variables that hold block by block alignment
 * info are big enough. */
if (blockAlloc < blockCount)
    {
    blockAlloc = blockCount*2;
    ExpandArray(oaStarts, 0, blockAlloc);
    ExpandArray(obStarts, 0, blockAlloc);
    ExpandArray(oBlockSizes, 0, blockAlloc);
    }


/* Get alignment from underlying psl and flip it as necessary so
 * that it become an alignment between the strand of a passed
 * in in "a->orientation" and whatever strand of b will stick.
 *
 * "a" can either be "q" or "t" in the psl, we don't know yet.
 * Inside psl "t" is always forward strand, "q" can vary.
 *
 * A further complication of this is that when "q" is 
 * reversed the coordinates of qStart and qEnd are in
 * terms of the forward strand, but the qStarts[i] and
 * qEnds[i] are in terms of the reverse strand, which
 * is related by the expression:
 *     qRevStart = qSize - qForEnd
 *     qRevEnd   = qSize - qForStart
 */

if (sameString(aFrag->name, psl->qName))
    {
    if (aOrient > 0)
	{
	if (edge->orientation > 0)
	    {
	    /* Straight forward case. */
	    forwardCopyBlocks(blockCount, psl->blockSizes, oBlockSizes, 
	    	psl->qStarts, oaStarts, psl->tStarts, obStarts);
	    bOrient = 1;
	    fprintf(logFile, ">>passed 1<<\n");
	    }
	else
	    {
	    /* Q is flipped and corresponds to A which is forward.
	     * Here we have to flip both Q and T as we copy to A and B. */
	    reverseCopyBlocks(blockCount, aSize, bSize,
	    	psl->blockSizes, oBlockSizes, 
	    	psl->qStarts, oaStarts, psl->tStarts, obStarts);
	    bOrient = -1;
	    fprintf(logFile, ">>passed 2<<\n");
	    }
	}
    else   /* a->orientation < 0 */
	{
	if (edge->orientation > 0)
	    {
	    /* Q and T are forward, but A is flipped.
	     * Have to flip both Q and T as we copy to A and B */
	    reverseCopyBlocks(blockCount, aSize, bSize,
	    	psl->blockSizes, oBlockSizes, 
	    	psl->qStarts, oaStarts, psl->tStarts, obStarts);
	    bOrient = -1;
	    fprintf(logFile, ">>passed 3<<\n");
	    }
	else
	    {
	    /* Q is reversed, T forward
	     * A reversed     B forward */
	    forwardCopyBlocks(blockCount, psl->blockSizes, oBlockSizes, 
	    	psl->qStarts, oaStarts, psl->tStarts, obStarts);
	    bOrient = 1;
	    fprintf(logFile, ">>passed 4<<\n");
	    }
	}
    }
else  /* A == T */
    {
    if (aOrient > 0)
	{
	if (edge->orientation > 0)
	    {
	    /* T forward  Q forward
	     * A forward  B forward. */
	    forwardCopyBlocks(blockCount, psl->blockSizes, oBlockSizes, 
	    	psl->tStarts, oaStarts, psl->qStarts, obStarts);
	    bOrient = 1;
	    fprintf(logFile, ">>passed 5<<\n");
	    }
	else
	    {
	    /* T forward Q reversed
	     * A forward B reversed */
	    forwardCopyBlocks(blockCount, psl->blockSizes, oBlockSizes, 
	    	psl->tStarts, oaStarts, psl->qStarts, obStarts);
	    bOrient = -1;
	    fprintf(logFile, ">>passed 6<<\n");
	    }
	}
    else	/* aOrient < 0 */
	{
	if (edge->orientation > 0)
	    {
	    /* T forward  Q forward
	     * A reversed B reversed */
	    reverseCopyBlocks(blockCount, aSize, bSize,
	    	psl->blockSizes, oBlockSizes, 
	    	psl->tStarts, oaStarts, psl->qStarts, obStarts);
	    bOrient = -1;
	    fprintf(logFile, ">>passed 7<<\n");
	    }
	else
	    {
	    /* T forward  Q reversed
	     * A reversed B forward */
	    reverseCopyBlocks(blockCount, aSize, bSize,
	    	psl->blockSizes, oBlockSizes, 
	    	psl->tStarts, oaStarts, psl->qStarts, obStarts);
	    bOrient = 1;
	    fprintf(logFile, ">>passed 8<<\n");
	    }
	}
    }

aQual = fragSeqQual(aFrag);
bQual = fragSeqQual(bFrag);
fprintf(logFile, "aOrient %d bOrient %d aQual %d bQual %d\n",
     aOrient, bOrient, aQual, bQual);
moreFromA = (aQual > bQual);
if (atMiddle == crossExtreme) 
    moreFromA = !moreFromA;

if (atMiddle == crossMiddle)
    {
    cbIx = midBigBlock(oBlockSizes, blockCount);
    }
else
    {
    if (moreFromA)
	cbIx = lastBigBlock(oBlockSizes, blockCount);
    else
	cbIx = firstBigBlock(oBlockSizes, blockCount);
    }

cbSize = oBlockSizes[cbIx];
cbPt = cbSize/2;
if (atMiddle != crossMiddle)
    {
    if (cbPt > 250)
	cbPt = cbSize - 250;
    if (!moreFromA)
	{
	cbPt = cbSize - cbPt;
	}
    }
aCross = oaStarts[cbIx] + cbPt;
if (aOrient > 0)
    {
    a->fragEnd = aCross;
    }
else
    {
    a->fragStart = aSize - aCross;
    }
bCross = obStarts[cbIx] + cbPt;
if (bOrient > 0)
    {
    b->fragStart = bCross;
    b->fragEnd = bSize;
    }
else
    {
    b->fragStart = 0;
    b->fragEnd = bSize - bCross;
    }
b->orientation = bOrient;
b->frag = bFrag;
}

struct raftFragRef 
/* Reference to a raft frag. */
    {
    struct raftFragRef *next;  /* Next in list. */
    struct raftFrag *rf;     /* Ref to frag. */
    struct goldenPath *gp;   /* Gp node associated with this. */
    enum crossAt crossAt;    /* Where to look for crossover. */
    };

struct raftFragRef *makeRfrList(struct raftFrag **bestFrags, int raftSize)
/* Return list of raft frag refs based on bestFrags, condensing bases next
 * to each other that refer to a single rf into one ref. */
{
struct raftFragRef *rfr, *rfrList = NULL;
struct raftFrag *rf, *last = NULL;
int i;

for (i=0; i<raftSize; ++i)
    {
    rf = bestFrags[i];
    if (rf != last && rf != NULL)
	{
	AllocVar(rfr);
	rfr->rf = rf;
	slAddHead(&rfrList, rfr);
	last = rf;
	}
    }
slReverse(&rfrList);
return rfrList;
}

struct raftFragRef *rfrLastRef(struct raftFragRef *rfrList, struct raftFrag *rf)
/* Return last rfr in list that refers to rf. */
{
struct raftFragRef *rfr, *hit = NULL;

for (rfr = rfrList; rfr != NULL; rfr = rfr->next)
    {
    if (rfr->rf == rf) 
	hit = rfr;
    }
assert(hit != NULL);
return hit;
}

struct raftFragRef *cleanRfrList(struct raftFragRef *oldList)
/* If a fragment is completely enclosed by another fragment remove
 * it from the list. */
{
struct raftFragRef *rfr, *next;
struct raftFragRef *lastRef, *badRef, *badNext;
struct raftFragRef *newList = NULL;

for (rfr = oldList; rfr != NULL; rfr = next)
    {
    lastRef = rfrLastRef(rfr, rfr->rf);
    for (badRef = rfr; badRef != lastRef; badRef = badNext)
	{
	fprintf(logFile, "Cleaning out %s\n", badRef->rf->frag->name);
	badNext = badRef->next;
	freeMem(badRef);
	}
    next = lastRef->next;
    slAddHead(&newList, lastRef);
    }
slReverse(&newList);
return newList;
}

struct goldenPath *gpFromRf(struct raftFrag *rf)
/* Construct a "default" golden path element that goes entirely
 * through fragment. */
{
struct oogFrag *frag = rf->frag;
struct goldenPath *gp;
AllocVar(gp);
gp->frag = frag;
gp->fragStart = 0;
gp->fragEnd = frag->size;
gp->orientation = rf->orientation;
return gp;
}

struct oogEdge *findSimpleConnectingEdge(struct oogFrag *a, 
	struct oogFrag *b, struct raft *raft)
/* Find best connecting edge between a and b. */
{
struct oogEdge *edge;
struct edgeRef *er;
int bestScore = -0x3fffffff;
struct oogEdge *bestEdge = NULL;

fprintf(logFile, "findSimpleConnectingEdge %s %s in raft %d (%d edges)\n",
	a->name, b->name, raft->id, slCount(raft->refList));
for (er = raft->refList; er != NULL; er = er->next)
    {
    edge = er->edge;
    if ( (edge->aSide == a && edge->bSide == b) || (edge->aSide == b && edge->bSide == a))
	{
	if (edge->score > bestScore)
	    {
	    bestScore = edge->score;
	    bestEdge = edge;
	    }
	}
    }
return bestEdge;
}


struct goldenPath *pathThroughRfrList(struct raftFragRef *rfrList, 
	struct raft *raft)
/* Find a path that connects fragments on rfrList.  */
{
struct goldenPath *lastGp = NULL, *gp, *gpList = NULL;
struct raftFragRef *rfr, *nextRfr;
struct oogEdge *edge;
int startOff = 0;

rfr = rfrList;
rfr->gp = gpList = lastGp = gpFromRf(rfr->rf);
if (rfr->next == NULL)
    return gpList;	/* Easy case. */

fprintf(logFile, "pathThroughRfrList got %d frags through raft %d starting %s\n", 
	slCount(rfrList), raft->id, lastGp->frag->name);
fprintf(logFile, "  %s\n", rfr->rf->frag->name);
for (rfr = rfrList->next; rfr != NULL; rfr = rfr->next)
    {
    fprintf(logFile, "  %s\n", rfr->rf->frag->name);
    if ((edge = findSimpleConnectingEdge(lastGp->frag, rfr->rf->frag, raft)) != NULL)
	{
	rfr->gp = gp = gpFromRf(rfr->rf);
	findCrossover(edge, lastGp->frag, gp->frag, lastGp, gp, rfr->crossAt);
	fprintf(logFile, "  lastGp (%s %d %d of %d) gp (%s %d %d of %d)\n", 
		lastGp->frag->name, lastGp->fragStart, lastGp->fragEnd, lastGp->frag->size,
		gp->frag->name, gp->fragStart, gp->fragEnd, gp->frag->size);
	slAddHead(&gpList, gp);
	lastGp = gp;
	}
    else
	{
	slFreeList(&gpList);
	return NULL;
	}
    }
slReverse(&gpList);
for (gp = gpList; gp != NULL; gp = gp->next)
    dumpGp(gp, logFile);
return gpList;
}

boolean gotReversals(struct goldenPath *gpList)
/* Check golden path for reversals. */
{
struct goldenPath *gp;

for (gp = gpList; gp != NULL; gp = gp->next)
    {
    if (gp->fragEnd <= gp->fragStart)
	{
	return TRUE;
	}
    }
return FALSE;
}


struct goldenPath *fixReversals(struct raftFragRef *rfrList, 
	struct raft *raft)
/* Fix reversals on golden path. */
{
struct goldenPath *gp;
struct raftFragRef *lastRfr = NULL, *rfr, *nextRfr;
boolean extreme;
boolean fixedIt = FALSE;
enum xSide {left = 0, right = 1, both = 2} xSide;

fprintf(logFile, "fixReversals\n");
for (extreme = 0; extreme <= 1; ++extreme)
    {
    enum crossAt crossAt = (extreme ? crossMiddle : crossExtreme);
    for (xSide = left; xSide <= both; ++xSide)
	{
	for (rfr = rfrList; rfr != NULL; rfr = rfr->next)
	    rfr->crossAt = crossUsual;
	for (rfr = rfrList; rfr != NULL; rfr = nextRfr)
	    {
	    nextRfr = rfr->next;
	    gp = rfr->gp;
	    if (gp->fragEnd <= gp->fragStart)
		{
		if (xSide == left || xSide == both)
		    rfr->crossAt = crossAt;
		if (nextRfr != NULL && (xSide == right || xSide == both))
		    nextRfr->crossAt = crossAt;
		}
	    lastRfr = rfr;
	    }
	gp = pathThroughRfrList(rfrList, raft);
	if (gp != NULL && !gotReversals(gp))
	    {
	    fixedIt = TRUE;
	    break;
	    }
	}
    if (fixedIt)
	break;
    fprintf(logFile, "Taking extreme measures to prevent reversals\n");
    }
if (gp == NULL)
    errAbort("Couldn't fix reversals", raft->id);
return gp;
}

struct goldenPath *hardPathThroughRaft(struct raft *raft)
/* We can't find a raft the easy way, so we'll make up a graph
 * and hope to find *some* way through the raft. */
{
struct diGraph *dg = dgNew();
struct dgNode *startNode, *endNode, *node, *aNode, *bNode;
struct dgNodeRef *nodeRefList, *nodeRef;
struct raftFragRef *rfrList = NULL, *rfr;
int startPos = 0x3fffffff;
int endPos = -startPos;
struct raftFrag *rf;
int s, e;
struct edgeRef *er;
struct oogEdge *edge;
struct goldenPath *gp;

fprintf(logFile, "taking hardPathThroughRaft %d\n", raft->id);
for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    s = rf->raftOffset;
    e = s + rf->frag->size;
    node = dgAddNode(dg, rf->frag->name, rf);
    if (s < startPos)
	{
	startPos = s;
	startNode = node;
	}
    if (e > endPos)
	{
	endPos = e;
	endNode = node;
	}
    }
for (er = raft->refList; er != NULL; er = er->next)
    {
    edge = er->edge;
    aNode = dgFindNode(dg, edge->aSide->name);
    bNode = dgFindNode(dg, edge->bSide->name);
    dgConnect(dg, aNode, bNode);
    dgConnect(dg, bNode, aNode);
    }
if ((nodeRefList = dgFindPath(dg, startNode, endNode)) == NULL)
    {
    errAbort("Can't find path from %s to %s in raft %d", startNode->name,
	endNode->name, raft->id);
    }
for (nodeRef = nodeRefList; nodeRef != NULL; nodeRef = nodeRef->next)
    {
    fprintf(logFile, "  %s\n", nodeRef->node->name);
    AllocVar(rfr);
    rfr->rf = nodeRef->node->val;
    slAddHead(&rfrList, rfr);
    }
slReverse(&rfrList);
gp = pathThroughRfrList(rfrList, raft);
if (gp != NULL)
    {
    if (gotReversals(gp))
	{
	gp = fixReversals(rfrList, raft);
	}
    }
else
    {
    errAbort("Curiously couldn't find hard way through raft %d", raft->id);
    }
slFreeList(&rfrList);
slFreeList(&nodeRefList);
dgFree(&dg);
return gp;
}

void removeReversals(struct goldenPath **pGpList)
/* As a last resort - simply remove the reversals. */
{
struct goldenPath *newList = NULL, *gp, *next;

for (gp = *pGpList; gp != NULL; gp = next)
    {
    next = gp->next;
    if (gp->fragEnd > gp->fragStart)
	{
	slAddHead(&newList, gp);
	}
    else
	{
	warn("Reversal *still* in golden path %d %d at %s removing....", 
	    gp->fragStart, gp->fragEnd, gp->frag->name);
	}
    }
slReverse(&newList);
*pGpList = newList;
}

void checkGp(struct goldenPath **pGpList)
/* Check golden path for reversals. */
{
struct goldenPath *gp;
bool gotReversals = FALSE;

for (gp = *pGpList; gp != NULL; gp = gp->next)
    {
    if (gp->fragEnd <= gp->fragStart)
	{
	gotReversals = TRUE;
	}
    }
removeReversals(pGpList);
}


struct goldenPath *pathThroughRaft(struct raft *raft, 
	struct hash *fragHash)
/* Return the best way through raft given orientation. */
{
int raftSize;
struct raftFrag **bestFrag, *rf;
int *bestScore;
struct raftFragRef *rfrList = NULL, *rfr;
struct goldenPath *gp;

fprintf(logFile, "pathThroughRaft %d at %d orientation %d\n", 
	raft->id, raft->defaultPos, raft->orientation);
if (raft->orientation < 0)
    {
    flipRaftFrags(raft);
    raft->orientation = 1;
    }
if (raft->start != 0 || raft->end <= 0)
    {
    errAbort("unnormalized raft!");
    normalizeRaft(raft);
    }
raftSize = raft->end;
AllocArray(bestFrag, raftSize);
AllocArray(bestScore, raftSize);

for (rf = raft->fragList; rf != NULL; rf = rf->next)
    {
    struct oogFrag *frag = rf->frag;
    int raftStart = rf->raftOffset;
    int fragSize = frag->size;
    int raftEnd = raftStart + fragSize;
    int score = fragSeqQual(frag);
    int lScore;
    int start, end;
    int i;
    start = raftStart;
    end = raftEnd;
    assert(end <= raftSize);
    for (i=start; i<end; ++i)
	{
	if (bestScore[i] < score)
	    {
	    bestFrag[i] = rf;
	    bestScore[i] = score;
	    }
	}
    }
rfrList = makeRfrList(bestFrag, raftSize);
rfrList = cleanRfrList(rfrList);
freeMem(bestScore);
freeMem(bestFrag);
gp = pathThroughRfrList(rfrList, raft);
if (gp == NULL)
    gp = hardPathThroughRaft(raft);
else
    {
    if (gotReversals(gp))
	gp = fixReversals(rfrList, raft);
    }
checkGp(&gp);
return gp;
}

int goldenPathSize(struct goldenPath *gpList)
/* Return number of bases in golden path. */
{
int total = 0;
struct goldenPath *gp;
for (gp = gpList; gp != NULL; gp = gp->next)
    {
    total += gp->fragEnd - gp->fragStart;
    }
return total;
}


void saveGoldenPath(char *fileName, char *contigName, struct goldenPath *goldenPath)
/* Save out golden path to file. */
{
FILE *f = mustOpen(fileName, "w");
struct goldenPath *gp;
int cStart = 0;
int fragSize;
int ix = 0;
struct oogFrag *frag;

for (gp = goldenPath; gp != NULL; gp = gp->next)
    {
    fragSize = gp->fragEnd - gp->fragStart;
    fprintf(f, "%s\t", contigName);
    fprintf(f, "%d\t%d\t", cStart+1, cStart+fragSize);
    fprintf(f, "%d\t", ++ix);
    if ((frag = gp->frag) != NULL)
	{
	fprintf(f, "%s\t%s\t%d\t%d\t%c\n", "O", frag->name, 
		gp->fragStart+1, gp->fragEnd,
		(gp->orientation > 0 ? '+' : '-'));
	}
    else
	{
	fprintf(f, "N\t%d\t%s\t%s\n", gp->fragEnd,
		gapTypeNames[gp->gapType],
		bridgeTypeNames[gp->bridgeType]);
	}
    cStart += fragSize;
    }
fclose(f);
}

int faLineSize = 50;
static int faLinePos = 0;
int faTotalSize = 0;

int faCharOut(FILE *f, char c)
/* Put out character to FA file, adding CR if necessary. */
{
fputc(toupper(c), f);
++faTotalSize;
if (++faLinePos >= faLineSize)
    {
    fputc('\n', f);
    faLinePos = 0;
    }
}

int writeGpFa(FILE *f, char *ctgName, struct goldenPath *goldenPath)
/* Save out fa with golden path. */
{
struct goldenPath *gp;

faLinePos = faTotalSize = 0;
fprintf(f, ">%s\n", ctgName);
for (gp = goldenPath; gp != NULL; gp = gp->next)
    {
    struct oogFrag *frag = gp->frag;
    int size = gp->fragEnd - gp->fragStart;
    int i;
    if (frag != NULL)
	{
	struct dnaSeq *seq = frag->seq;
	DNA *dna = seq->dna + gp->fragStart;
	if (gp->orientation < 0)
	    reverseComplement(dna, size);
	for (i=0; i<size; ++i)
	    faCharOut(f, dna[i]);
	if (gp->orientation < 0)
	    reverseComplement(dna, size);
	}
    else
	{
	for (i=0; i<size; ++i)
	    {
	    faCharOut(f, 'N');
	    }
	}
    }
if (faLinePos != 0)
    fputc('\n', f);
return faTotalSize;
}

void makeGoldenFa(char *faOutName, char *ctgName, struct goldenPath *goldenPath)
/* Save out fa with golden path. */
{
FILE *f = mustOpen(faOutName, "w");
int faTotalSize = writeGpFa(f, ctgName, goldenPath);
fclose(f);
printf("Wrote %d total bases to %s\n", faTotalSize, faOutName);
}

void makeCopperFa(char *fileName, struct raftPath *raftPath)
/* Write out fa file with rafts as separate records, no N's */
{
FILE *f = mustOpen(fileName, "w");
struct raftPath *rp;
char name[16];

for (rp = raftPath; rp != NULL; rp = rp->next)
    {
    sprintf(name, "raft%d", rp->raft->id);
    writeGpFa(f, name, rp->gpList);
    }
fclose(f);
}

void saveGl(char *glName, struct raftPath *raftPath)
/* Save layout as a .gl file. */
{
struct raftPath *rp;
int raftOffset;
struct raft *raft;
struct raftFrag *rf;
struct oogFrag *frag;
FILE *f = mustOpen(glName, "w");
int start, end;

for (rp = raftPath; rp != NULL; rp = rp->next)
    {
    raft = rp->raft;
    raftOffset = rp->start;
    slSort(&raft->fragList, raftFragCmpOffset);
    for (rf = raft->fragList; rf != NULL; rf = rf->next)
	{
	frag = rf->frag;
	start = raftOffset + rf->raftOffset;
	end = start + frag->size;
	fprintf(f, "%s %d %d %c\n",
	    frag->name, start, end,
	    (rf->orientation >= 0 ? '+' : '-'));
	}
    }
fclose(f);
}

boolean realConnection(struct dgConnection *conList)
/* Return TRUE if there's a connection with val (not just
 * a dummy link to clone ends). */
{
struct dgConnection *con;
for (con = conList; con != NULL; con = con->next)
    {
    if (con->node->val != NULL)
	return TRUE;
    }
return FALSE;
}

boolean inGraph(struct diGraph *graph, struct raft *raft)
/* Return TRUE if raft is connected inside graph. */
{
struct dgNode *node;
node = dgFindNumberedNode(graph, raft->id);
return realConnection(node->nextList) || realConnection(node->prevList);
}

void saveLayout(char *glName, struct raftPath *raftPath, struct diGraph *graph)
/* Save layout as a gl.N file. */
{
struct raftPath *rp;
int raftOffset;
struct raft *raft;
struct raftFrag *rf;
struct oogFrag *frag;
FILE *f = mustOpen(glName, "w");
int start, end;

for (rp = raftPath; rp != NULL; rp = rp->next)
    {
    raft = rp->raft;
    raftOffset = rp->start;
    slSort(&raft->fragList, raftFragCmpOffset);
    for (rf = raft->fragList; rf != NULL; rf = rf->next)
	{
	frag = rf->frag;
	start = raftOffset + rf->raftOffset;
	end = start + frag->size;
	fprintf(f, "%s\t%d\t%d\t%c\t",
	    frag->name, start, end,
	    (rf->orientation >= 0 ? '+' : '-'));
	if (inGraph(graph, raft))
	    {
	    fprintf(f, "G%d\t", raft->node->component);
	    }
	else
	    {
	    fprintf(f, " \t");
	    }
	if (slCount(raft->fragList) > 1)
	    {
	    fprintf(f, "R%d\n", raft->id);
	    }
	else
	    {
	    fprintf(f, " \n");
	    }
	}
    }
fclose(f);
}

struct goldenPath *gapInGp(struct raftPath *a, struct raftPath *b, struct diGraph *bargeGraph)
/* Make up data structure to describe gap between a and b. */
{
struct goldenPath *gp;
struct raft *aRaft = a->raft, *bRaft = b->raft;
enum gapType gt;
enum bridgeType bt;
int nCount;
struct barge *aBarge = aRaft->barge, *bBarge = bRaft->barge;

if (aBarge == bBarge)
    {
    gt = gtFragment;
    nCount = fragGapSize;
    if (aRaft->node->component == bRaft->node->component)
	bt = btYes;
    else
	bt = btNo;
    }
else
    {
    gt = gtClone;
    if (dgDirectlyFollows(bargeGraph, aBarge->graphNode, bBarge->graphNode))
	{
	bt = btYes;
	nCount = bridgedBargeGapSize;
	}
    else
	{
	bt = btNo;
	nCount = unbridgedBargeGapSize;
	}
    }
AllocVar(gp);
gp->fragEnd = nCount;
gp->gapType = gt;
gp->bridgeType = bt;
return gp;
}

void edgesIntoBarges(struct barge *bargeList, struct oogEdge **pEdgeList)
/* edgesIntoBarges - move edges from master edge list to barge-specific
 * edge lists.  Leave edges that cross between barges on *pEdgeList. */
{
struct oogEdge *oldList = *pEdgeList, *bargeCrossers = NULL, *edge, *nextEdge;
struct barge *barge;
struct oogClone *clone;
int crossCount = 0;

for (edge = oldList; edge != NULL; edge = nextEdge)
    {
    nextEdge = edge->next;
    barge = edge->aSide->clone->barge;
    assert(barge != NULL);
    if (barge == edge->bSide->clone->barge)
	{
	slAddHead(&barge->selfEdges, edge);
	}
    else
	{
	slAddHead(&bargeCrossers, edge);
	++crossCount;
	}
    }
status("eliminated %d raft edges that cross barges\n", crossCount);
for (barge = bargeList; barge != NULL; barge = barge->next)
    slReverse(&barge->selfEdges);
slReverse(&bargeCrossers);
*pEdgeList = bargeCrossers;
}

void ooGreedy(char *inDir)
/* Try and do layouts specified in inDir and put results in
 * outFile. */
{
static char mrnaPslName[512];
static char estPslName[512];
static char selfPslName[512];
static char raftPslName[512];
static char oogLogName[512];
static char raftOutName[512];
static char bargeOutName[512];
static char graphOutName[512];
static char infoName[512];
static char genoListName[512];
static char fragMapName[512];
static char raftOrderName[512];
static char goldName[512];
static char faOutName[512];
static char bacPairName[512];
static char bacPslName[512];
static char glName[512];
static char layoutName[512];
struct oogEdge *selfEdgeList = NULL;
struct mrnaLine *mrnaLineList = NULL, *ml;
struct mrnaLine *estLineList = NULL;
struct mrnaLine *bacEndLineList = NULL;
struct oogClone *cloneList = NULL;
struct oogFrag *fragList = NULL;
struct hash *cloneHash = newHash(8);
struct hash *fragHash = newHash(12);
struct hash *ocpHash = newHash(0);
struct hash *coHash = NULL;
struct clonePair *coList = NULL;
struct raft *raftList = NULL, *raft, *oneRaftSet;
struct barge *bargeList, *barge;
struct psl *selfPslList = NULL;
FILE *graphOut; /* File to write graph to. */
FILE *raftOut = NULL;  /* File to write raft to. */
int raftIx;
struct diGraph *raftGraph;
struct diGraph *bargeGraph;
int contigTotalSize = 0;
int raftTotalSize = 0;
int contigPos = 0;
int raftCount = 0;
struct dgNodeRef *orderList, *orderEl;
struct goldenPath *goldenPath = NULL, *gp;
struct raftPath *raftPath = NULL, *rp;
char *ctgName;

/* Make up a bunch of file names based on target directory and
 * current version number of program. */
sprintf(selfPslName, "%s/self.psl", inDir);
sprintf(mrnaPslName, "%s/mrna.psl", inDir);
sprintf(estPslName, "%s/est.psl", inDir);
sprintf(raftPslName, "%s/raft.psl", inDir);
sprintf(oogLogName, "%s/ooGreedy.log", inDir, version);
sprintf(raftOutName, "%s/raft.%d", inDir, version);
sprintf(bargeOutName, "%s/barge.%d", inDir, version);
sprintf(graphOutName, "%s/graph.%d", inDir, version);
sprintf(infoName, "%s/info", inDir);
sprintf(fragMapName, "%s/fragMap.%d", inDir, version);
sprintf(raftOrderName, "%s/raftOrder.%d", inDir, version);
sprintf(genoListName, "%s/geno.lst", inDir);
sprintf(goldName, "%s/gold.%d", inDir, version);
sprintf(bacPairName, "%s/bacEndPairs", inDir);
sprintf(bacPslName, "%s/bacEnd.psl", inDir);
sprintf(glName, "%s/ooGreedy.%d.gl", inDir, version);
sprintf(layoutName, "%s/gl.%d", inDir, version);
sprintf(looseFileName, "%s/loose.%d", inDir, version);
sprintf(tightFileName, "%s/tight.%d", inDir, version);

/* Open log file file. */
logFile = mustOpen(oogLogName, "w");
pushWarnHandler(warnHandler);

status("ooGreedy version %d on %s\n", version, inDir);

/* Load in DNA and make clone and fragment lists. */
contigTotalSize =  loadAllClones(infoName, genoListName, fragHash, 
	cloneHash, &fragList, &cloneList, &ctgName);
sprintf(faOutName, "%s/%s.fa", inDir, ctgName);
status( "Got %d clones %d frags ~%d contig size\n", 
	slCount(cloneList), slCount(fragList), contigTotalSize);

/* Load up self data structures and sort edge list. */
processSelf(selfPslName, &cloneList, &fragList, &selfEdgeList, &selfPslList, cloneHash, fragHash);
status("Found %d fragment overlaps in self.psl\n", slCount(selfEdgeList));

#ifdef OLD
/* Make initial rafts from which we will derive barges. */
slSort(&selfEdgeList, oogEdgeCmp);
raftList = makeRafts(fragList, selfEdgeList);
status("found %d initial rafts (%d of %d frags) in %s\n", 
	slCount(raftList), raftedFragCount(fragList),
	slCount(fragList), selfPslName);
#endif /* OLD */

/* Make barges. */
bargeList = makeBarges(cloneList, raftList, selfPslList, fragHash, cloneHash, ocpHash);
status(" %d barges in %d clones\n", slCount(bargeList), slCount(cloneList));

/* Remake rafts forcing them to fit on barges. */
#ifdef OLD
unlinkFragsFromRafts(fragList);
raftFreeList(&raftList);
#endif /* OLD */

edgesIntoBarges(bargeList, &selfEdgeList);
uglyf("Past edges into barges\n");
raftOut = mustOpen(raftOutName, "w");
raftPsl = mustOpen(raftPslName, "w");
pslWriteHead(raftPsl);
for (barge = bargeList; barge != NULL; barge = barge->next)
    {
    uglyf("About to makeRafts on %d edges\n", slCount(barge->selfEdges));
    oneRaftSet = makeRafts(fragList, barge->selfEdges);
    uglyf("About to slCat %x %x\n", raftList, oneRaftSet);
    raftList = slCat(raftList, oneRaftSet);
    uglyf("Past slCat\n");
    }
uglyf("past loop B\n");    
fprintf(raftOut, "ooGreedy version %d found %d rafts (%d of %d frags) in %s\n\n", 
	version, slCount(raftList), raftedFragCount(fragList),
	slCount(fragList), selfPslName);
status("found %d rafts (%d of %d frags) in %s\n", 
	slCount(raftList), raftedFragCount(fragList),
	slCount(fragList), selfPslName);

makeSingletonRafts(fragList, &raftList);
calcPositions(bargeList, fragList, NULL);
linkRaftsToBarges(raftList);


/* Read in rna edge lists.  Merge est edges into rna ones. */
    {
    int mCount = 0, eCount = 0, bCount = 0;
    mrnaLineList = makeMrnaLines(mrnaPslName, fragHash);
    mCount = slCount(mrnaLineList);
    estLineList = makeMrnaLines(estPslName, fragHash);
    eCount = slCount(estLineList);
    mrnaLineList = slCat(mrnaLineList, estLineList);
    estLineList = NULL;
    bacEndLineList = makeBacEndLines(bacPairName, bacPslName, fragHash);
    bCount = slCount(bacEndLineList);
    mrnaLineList = slCat(mrnaLineList, bacEndLineList);
    bacEndLineList = NULL;
    slSort(&mrnaLineList, mrnaLineCmp);
    fprintf(stdout, "Got %d mRNA, %d EST, and %d BAC end pair bridges\n",
	mCount, eCount, bCount);
    fprintf(logFile, "Got %d mRNA, %d EST, and %d BAC end pair bridges\n",
	mCount, eCount, bCount);
    }


fprintf(logFile, "\nList of mrnaLines:\n");
for (ml = mrnaLineList; ml != NULL; ml = ml->next)
    dumpMrnaLine(ml, logFile);
makeGraphs(raftList, bargeList, mrnaLineList, &raftGraph, &bargeGraph);
status("Made bridge graphs\n");

calcPositions(bargeList, fragList, bargeGraph);
setRaftDefaultPositions(raftList);
slSort(&fragList, cmpFragDefaultPos);
writeFragMap(fragMapName, fragList);
for (raft = raftList; raft != NULL; raft = raft->next)
    {
    normalizeRaft(raft);
    dumpRaft(raft, raftOut);
    fprintf(raftOut, "\n");
    }
carefulClose(&raftOut);
status("Updated default positions\n");


flipNearDefaults(raftGraph);
graphOut = mustOpen(graphOutName, "w");
fprintf(graphOut, "mRNA/BAC ends graph by ooGreedy version %d\n\n", version); 
dgConnectedComponentsWithVals(raftGraph);
ooDumpGraph(raftGraph, graphOut);
carefulClose(&graphOut);
status("Flipped and saved graphs.\n");

saveBarges(bargeOutName, bargeList, bargeGraph, ocpHash);

orderList = orderRafts(raftGraph);
status("Ordered rafts\n");

	{
	int count = 0;
	fprintf(logFile, "Order of rafts:\n");
	fprintf(logFile, "\t");
	for (orderEl = orderList; orderEl != NULL; orderEl = orderEl->next)
	    {
	    raft = orderEl->node->val;
	    if (raft != NULL)
		{
		fprintf(logFile, "%d ", raft->id);
		if (++count % 10 == 0)
		    fprintf(logFile,"\n\t");
		}
	    }
	fprintf(logFile, "\n");
	}

/* Find path through each raft. */
for (orderEl = orderList; orderEl != NULL; orderEl = orderEl->next)
    {
    raft = orderEl->node->val;
    if (raft != NULL)
	{
	AllocVar(rp);
	rp->gpList = pathThroughRaft(raft, fragHash);
	rp->raft = raft;
	rp->size = goldenPathSize(rp->gpList);
	raftTotalSize += rp->size;
	++raftCount;
	slAddHead(&raftPath, rp);
	}
    }
slReverse(&raftPath);

	{
	char fileName[512];
	status("Writing copper path\n");
	sprintf(fileName, "%s/copper.fa", inDir);
	makeCopperFa(fileName, raftPath);
	}

/* Merge together raft paths into full path */
for (rp = raftPath,raftIx = 0; rp != NULL; rp = rp->next, ++raftIx)
    {
    goldenPath = slCat(goldenPath, rp->gpList);
    rp->gpList = NULL;
    rp->start = contigPos;
    contigPos += rp->size;
    if (rp->next != NULL)
	{
	gp = gapInGp(rp, rp->next, bargeGraph);
	contigPos += gp->fragEnd;
	slAddTail(&goldenPath, gp);
	}
    }
status("Found golden path\n");

saveGoldenPath(goldName, ctgName, goldenPath);
makeGoldenFa(faOutName, ctgName, goldenPath);
saveGl(glName, raftPath);
saveLayout(layoutName, raftPath, raftGraph);
status("\n");
}

int main(int argc, char *argv[])
/* Process command line. */
{
if (argc != 2)
    usage();
ooGreedy(argv[1]);
#ifdef SOMETIMES
ooGreedy("easy");
#endif
return 0;
}
